# 
# AnImAl support functions
#
ia.size <- function (img) {

  return(dim(img@data)[2] * dim(img@data)[1])
  
}

ia.mask <- function (img) {
  
  if(is.animage(img)) {
    animask(img@focus[1],
            img@focus[2],
            dim(img@data)[2],
            dim(img@data)[1])
  } else {
    NULL
  }
  
}

ia.maskLocator <- function (img) {

  imgDims <- dim(img@data)
  bounds  <- locator(2, type="l")
  
  x0 <- min(bounds$x)
  y0 <- dim(img@data)[1] -  max(bounds$y)
  dx <- max(bounds$x) - x0
  dy <- max(bounds$y) - min(bounds$y)

  animask(as.integer(x0),
          as.integer(y0),
          as.integer(dx),
          as.integer(dy))
  
}

ia.frame <- function(img, frame=1L) {
  mask <- ia.mask(img)

  mask@x <- mask@x + as.integer(frame)
  mask@y <- mask@y + as.integer(frame)
  mask@w <- mask@w - as.integer(2*frame)
  mask@h <- mask@h - as.integer(2*frame)

  ia.get(img, mask)
}

ia.xmin <- function (img) {
  
  if(is.animage(img)) {
    return(img@focus[1])
  } else {
    NULL
  }
  
}

ia.xmax <- function (img) {
  
  if(is.animage(img)) {
    return(img@focus[1] + dim(img@data)[2] - 1)
  } else {
    NULL
  }
  
}

ia.ymin <- function (img) {
  
  if(is.animage(img)) {
    return(img@focus[2])
  } else {
    NULL
  }
  
}

ia.ymax <- function (img) {
  
  if(is.animage(img)) {
    return(img@focus[2] + dim(img@data)[1] - 1)
  } else {
    NULL
  }
  
}

ia.updatedHistory <- function (callinfo) {

  i <- 0
  for(name in names(callinfo)) {
    i <- i + 1
    if(nchar(name) > 0) {
      if(is.animage(eval(as.symbol(name), parent.frame(1)))) {
        if(!(eval(as.symbol(name), parent.frame(1)))@document) {
          callinfo <- FALSE
          break;
        } else {
          callinfo[[i]] <- eval(as.symbol(name), parent.frame(1))@history
        }
      } else if (is.animageList(eval(as.symbol(name), parent.frame(1)))){
        if(!(is.documentedAnimageList(eval(as.symbol(name), parent.frame(1))))) {
          callinfo <- FALSE
          break;
        } else {
          listHistory <- call("list")
          ih          <- 1
          for(img in eval(as.symbol(name), parent.frame(1))) {
            ih                <- ih + 1
            listHistory[[ih]] <- img@history
          }
          callinfo[[i]] <- listHistory
        }
      } else {
        callinfo[[i]] <- eval(as.symbol(name), parent.frame(1))
      }
    }
  }
  
  callinfo
  
}

#
# ia.get works correctly on a list of images but history is not
# managed in this case
#
ia.get <- function (img, locus) {

  if(is.list(img)) {
    res <- ia.map(function(i) ia.get(i, locus), img)
  } else {
    res <- aniGet(img, locus)
  }

  if(is.animage(res)) {
    if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
      res@history  <- updatedHistory
      res@document <- TRUE
    } else {
      res@history  <- call("<undef>")
      res@document <- FALSE
    }
  }
  
  res

}

ia.up <- function (img, delta = 1L) {

  focus    <- img@focus
  focus[2] <- focus[2] - delta
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = focus)
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.down <- function (img, delta = 1L) {

  focus    <- img@focus
  focus[2] <- focus[2] + delta
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = focus)
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.left <- function (img, delta = 1L) {

  focus <- img@focus
  
  focus[1] <- focus[1] - delta
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = as.integer(focus))
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.right <- function (img, delta = 1L) {

  focus    <- img@focus
  focus[1] <- focus[1] + delta
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = focus)
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.slide <- function (img, dx = 1L, dy = 1L) {

  focus    <- img@focus
  focus[1] <- focus[1] + as.integer(dx)
  focus[2] <- focus[2] + as.integer(dy)
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = focus)
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.translateAt <- function (img, at = c(0L, 0L)) {

  focus    <- img@focus
  focus[1] <- at[1]
  focus[2] <- at[2]
  res      <- animage(data     = img@data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = focus)
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
  
}

ia.packBitImages <- function(intAnimageList, thr = 128L) {
  
  if(length(intAnimageList) > as.integer(round(log2(.Machine$integer.max))))
    stop("too many images for packing")
  
                                        # compute documentation
                                        # status and inhibit for
                                        # internal computations
  document <- c()
  
  for(i in 1:length(intAnimageList))  {
    document[[i]]        <- intAnimageList[[i]]@document
    intAnimageList[[i]]@document <- FALSE
  }

  res          <- ia.greater(intAnimageList[[1]], thr)

  for(i in 2:length(intAnimageList)) {
    res <- ia.or(res,
                 ia.shift(ia.greater(intAnimageList[[i]], thr),
                          as.integer((i-1))))
  }

  for(i in 1:length(intAnimageList))  
    intAnimageList[[i]]@document <- document[[i]]
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  
  res

}

ia.unpackBitImages <- function(img, planes = 8L) {

  res          <- c()
  document     <- img@document
  img@document <- FALSE
  
  for(p in 1:planes) {
    res    <- c(res, ia.and(img, 1L))
    img    <- ia.shift(img, -1L)
  }
  
  for(p in 1:planes) 
    res[[p]]@document <- document
        
  res

}

ia.integerAnimage <- function(img) {
  
  res <- new("animage",
             data     = array(as.integer(img@data), dim(img@data)),
             pmode    = img@pmode,
             focus    = img@focus,
             outside  = as.integer(img@outside),
             storage  = "integer",
             document = img@document)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  
  res
}

ia.realAnimage <- function(img) {
  
  res <- new("animage",
             data     = array(as.real(img@data), dim(img@data)),
             pmode    = img@pmode,
             focus    = img@focus,
             outside  = as.real(img@outside),
             storage  = "real",
             document = img@document)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  
  res
}

#
# Normalize the scale of image values
#

ia.scale <- function(A, maxValue = 1.0) {
  if(is.animage(A))
    {
      return(ia.mult(A, maxValue / max(A@data)))
    }
  else
    {
      return(lapply(A,
                    ia.scale,
                    maxValue))
    }
}

#
# A (and optionally B) may be an animage or a list of animages
#

ia.congruent <- function(A, B = FALSE) {

  if(is.animage(A) && is.animage(B))
    {
      dimA <- dim(A@data)
      dimB <- dim(B@data)
      atA  <- A@focus
      atB  <- B@focus
      
    if((A@storage == B@storage) &&
       (atA[1]    == atB[1])    &&
       (atA[2]    == atB[2])    &&
       (dimA[1]   == dimB[1])   &&
       (dimA[2]   == dimB[2])) {
      return(TRUE);
    } else {
      return(FALSE);
    }
  } else {
    congruent <- TRUE

    if(is.animage(A))
      A <- c(A)
    else
      if(length(A) > 1)
        for(a in 2:length(A))
          congruent <- (congruent && ia.congruent(A[[1]], A[[a]]))

    
    if(is.animage(B))
      B <- c(B)

    if(is.animage(B[[1]]))
      for(b in 1:length(B))
        congruent <- (congruent && ia.congruent(A[[1]], B[[b]]))
    
    return(congruent)
  }
  
}

ia.statImage <- function(img) {
  m  <- ia.mask(img)
  yu <- m@y
  xl <- m@x
  xr <- xl + m@w -1
  yd <- yu + m@h -1

  list(xl, xr, yu, yd)
}


ia.apply <- function(img, f) {
  
  data <- img@data
  d    <- dim(data)
  
  jmin <- 1
  jmax <- d[1]

  imin <- 1
  imax <- d[2]

  xmin <- img@focus[1]
  ymin <- img@focus[2]

  x    <- xmin
  for(i in imin:imax) {
    y <- ymin
    for(j in jmin:jmax) {
      data[j,i] <- f(x, y, data[j,i])
      y         <- y+1
    }
    x <- x+1
  }

  res <- new("animage",
             data     = data,
             pmode    = img@pmode,
             focus    = img@focus,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  
  res 
}

#
# A recursive definition of the folding operation
# reduce
#
ia.reduce <- function(op,X) {
  op <- match.fun(op)
  n  <- length(X)
  
  if (n<2)
    stop("need n>1")

  if(is.list(X))
    if(n==2)
      op(X[[1]],X[[2]])
    else
      op(X[[1]], ia.reduce(op,X[-1]))  
  else if (is.vector(X))
    if(n==2)
      op(X[1],X[2])
    else
      op(X[1], ia.reduce(op,X[-1]))
}

#
# A wrapper to capture lisp mapcar functionalities
#
ia.map <- function(op, ...) {
  mapply(op, ..., SIMPLIFY = FALSE)
}

ia.subsample2DMatrix <- function(mat, factor=2) {
  rows <- dim(mat)[[1]]
  cols <- dim(mat)[[2]]

  if(length(factor) == 1)
    mat[seq(1,rows,by=factor),seq(1,cols,by=factor)]
  else
    mat[seq(1,rows,by=factor[[1]]),seq(1,cols,by=factor[[2]])]
  
}

ia.subsample <- function(img, factor=2) {
  res      <- animage(data     = ia.subsample2DMatrix(img@data, factor),
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = (img@focus)/factor)

    if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
}

