#
#
#
ia.correlation <- function(img,
                           template,
                           dd        =  0L,
                           type      =  "Pearson",
                           normalize =  TRUE,
                           p         =  2L,
                           range     =  1,
                           filter    =  FALSE,
                           mask      =  FALSE)
{
  if(!is.animage(img)  || (img@storage != "real"))
    stop("<img> must be a real animage");
  
  if(! is.animage(template)  || (template@storage != "real"))
    stop("<template> argument must be an animage");

  if(!is.integer(dd))
    stop("<dd> neighborhood specification must be integer");

  if(type == "Pearson")
    type = 0L
  else if (type == "G")
    type = 1L
  else if (type == "L")
    type = 2L
  else if (type == "Lp")
    type = 3L
  else
    stop("illegal correlation <type> selector")
  
  if(!is.integer(p) || (p <= 0L))
    stop("<p> must be non negative integer")

  if((is.logical(filter) && (filter != FALSE)) ||
     (is.animage(filter) && (filter@storage != "real")))
    stop("<filter> must be FALSE or a real animage");

  if(!(mask == FALSE) || (is.animage(mask) && (mask@storage != "real")))
    stop("<mask> must be FALSE or a real animage");

  #
  # if L_p matching is required, we need to compute
  # a proper normalization factor
  #
  if(type == 3) {
    maxD <- range
  } else {
    maxD <- 0
  }

  if(normalize == TRUE)
    norm = 1L
  else
    norm = 0L

  res <- .Call("aniCorrelation",
               img,
               template,
               dd,
               type,
               norm,
               p,
               maxD,
               filter,
               mask,
               PACKAGE="AnImAl")
  
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res[[1]]@history  <- updatedHistory
    res[[1]]@document <- TRUE
  } else {
    res[[1]]@history  <- call("<undef>")
    res[[1]]@document <- FALSE
  }

  res
}
