#
#
#
ia.moravec <- function(img, width=3L) {
  #
  document     <- img@document
  img@document <- FALSE
  #
  # The set of directions along which
  # image differences are computed
  #
  dirs <- list(c( 1, 0),
               c( 1, 1),
               c( 0, 1),
               c(-1, 1))
  #
  # given a direction, compute the squared difference
  # along it and integrate over a window of specified
  # size
  #
  f <- function(a) {
    delta = ia.sub(img, ia.slide(img, -a[1], -a[2]))
    ia.average(ia.mult(delta, delta), width, width, norm = 0L)
  }
  #
  # compute the result based on the full set of directions
  #
  res <- ia.get(ia.reduce(ia.min,
                          ia.map(f, dirs)),
                ia.mask(img))
  #
  # get rid of boundary artifacts
  #
  res@outside <- 0
  res         <- ia.frame(ia.frame(res, width), -width)
  #
  # manage history tracing
  #
  img@document <- document
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  #
  res
}
