ia.mmBitDilation <- function(img, strctElt) {
  idxs <- which(strctElt@data > 0L, arr.ind=TRUE) + strctElt@focus - c(1L, 1L)
  n    <- dim(idxs)[1]
  at   <- img@focus
  
  dImg <- ia.translateAt(img, at+idxs[1,])

  for(i in 2:n) 
    dImg <- ia.or(dImg, ia.translateAt(img, at+idxs[i,]))

  ia.get(dImg, ia.mask(img))
}

ia.mmBitErosion <- function(img, strctElt) {
  idxs <- which(strctElt@data > 0, arr.ind=TRUE) + strctElt@focus - c(1L, 1L)
  n    <- dim(idxs)[1]

  at   <- img@focus
  dImg <- ia.translateAt(img, at+idxs[1,])

  for(i in 2:n) 
    dImg <- ia.and(dImg, ia.translateAt(img, at+idxs[i,]))

  ia.get(dImg, ia.mask(img))
}

ia.mmGreyDilation <- function(img, strctElt) {
  xs = strctElt@focus[1]
  ys = strctElt@focus[2]
  xe = xs + dim(strctElt@data)[2]-1
  ye = ys + dim(strctElt@data)[1]-1
  
  at   <- img@focus
  dImg <- ia.add(ia.translateAt(img, at+c(xs,ys)),
                 strctElt[ys,xs])
  
  for(y in ys:ye) {
    for(x in xs:xe) {
      if(!((x == xs) && (y == ys)))
        dImg <- ia.max(dImg,
                       ia.add(ia.translateAt(img, at+c(x,y)),
                              strctElt[y,x]))
    }
  }
  
  ia.get(dImg, ia.mask(img))
}

ia.mmGreyErosion <- function(img, strctElt) {
  xs = strctElt@focus[1]
  ys = strctElt@focus[2]
  xe = xs + dim(strctElt@data)[2]-1
  ye = ys + dim(strctElt@data)[1]-1
  
  at   <- img@focus
  dImg <- ia.sub(ia.translateAt(img, at+c(xs,ys)),
                 strctElt[ys,xs])
  
  for(y in ys:ye) {
    for(x in xs:xe) {
      if(!((x == xs) && (y == ys)))
        dImg <- ia.min(dImg,
                       ia.sub(ia.translateAt(img, at+c(x,y)),
                              strctElt[y,x]))
    }
  }
  
  ia.get(dImg, ia.mask(img))
}

ia.mmBitOpening <- function(img, strctElt) {
  ia.mmBitDilation(ia.mmBitErosion(img, strctElt), strctElt)
}

ia.mmBitClosing <- function(img, strctElt) {
  ia.mmBitErosion(ia.mmBitDilation(img, strctElt), strctElt)
}

ia.mmGreyOpening <- function(img, strctElt) {
  ia.mmGreyDilation(ia.mmGreyErosion(img, strctElt), strctElt)
}

ia.mmGreyClosing <- function(img, strctElt) {
  ia.mmGreyErosion(ia.mmGreyDilation(img, strctElt), strctElt)
}

ia.mmTopHatTransform <- function(img, strctElt, mode="white") {
  if(mode == "white") {
    ia.sub(img,
           ia.mmGreyOpening(img, strctElt))
  } else if (mode == "black") {
    ia.sub(ia.mmGreyClosing(img, strctElt),
           img)
  } else {
    stop("illegal mode value: should be white/black")
  }
}
