require(AnImAl)

tiffdir <- file.path(system.file(package = "AnImAl"), "sampleimages")

# This is a more awkward way to get the same info
#
# tifffiles <- (function (x) { file.path(tiffdir, x) }) (dir(pattern = glob2rx("*.tif")))
#
tiffiles  <- list.files(tiffdir, pattern = glob2rx("*.tif"))

opar  <- par(ask = dev.interactive(orNone = TRUE))

##
## Read a standard TIFF file with alpha channel
##

sampleFace <- ia.readAnimage(file.path(tiffdir, "sampleFace_06.tif"))

ia.show(sampleFace)

##
## Read a HDR TIFF file in float format with alpha channel
##
hdr   <- ia.readAnimage(file.path(tiffdir, "hdr_float.tif"))

##
## as you can see the dynamic range is awkward
##
hist(hdr[[1]]@data)

##
## this can be better appreciated in log scale
##
hist(log(hdr[[1]]@data))

##
## we then need to gamma correct the image
##
gammaCorrection <- function(x) { return (x^(0.4545)) }

plot(pixmapRGB(abind(gammaCorrection(hdr[[1]]@data),
                     gammaCorrection(hdr[[2]]@data),
                     gammaCorrection(hdr[[3]]@data),
                     rev.along=0)))

##
## ... more of the same
##
gammaCorrection <- function(x) { return (x^(0.3)) }

plot(pixmapRGB(abind(gammaCorrection(hdr[[1]]@data),
                     gammaCorrection(hdr[[2]]@data),
                     gammaCorrection(hdr[[3]]@data),
                     rev.along=0)))

#
# Simple image transformation using 'aniTable':
#
# invert of the green channel
#
inverted <- ia.table(sampleFace[[2]], 255:0, 0L)
#
# and visualize it
#
ia.show(inverted)

par(opar)

##
## END
##
