\name{animage-class}
\docType{class}
\alias{animage-class}
\alias{animage}
\alias{as.animage}
\alias{as.animage,animage-method}
\alias{as.animage,pixmapRGB-method}
\alias{as.animage,array-method}
\alias{show,animage-method}
\alias{plot,animage-method}
\alias{print.animage}
\alias{hist.animage}
\alias{is.animage}
\alias{is.animageList}
\alias{is.documentedAnimageList}
\alias{[,animage-method}
\alias{[<-,animage-method}
\title{Class "animage"}

\description{A simple image abstraction that supports
  self-documentation and region of interest localized processing.
  }

\section{Objects from the Class}{

  Objects can be created by calls of the form \code{new("animage", ...)}.
These objects can represent images with two storage types (integer or
real) and can be considered as infinite images with scalar pixel values
defined by a (possibly) non-costant rectangular set of pixels surrounded
by an infinite area of constant value.


  \code{animage} is the preferred constuctor for \code{animage} objects

  \code{as.animage} coerces an array object to an \code{animage} object

  \code{is.animage,is.animageList,is.documentedAnimageList} checks for
  their argument to be respectively, an \code{animage}, a list of
  \code{animage} objects,  a list of
  \code{animage} objects with \code{document = TRUE}
  
  Implementation via S4  methods has been opted for  in part to simplify
C-language interface,  providing stricter type  checking, and convenient
mirroring of R level slots into C-language accessors.  In order to get a
more compact  syntax, tests  for class, coercion,  and print rely  on S3
syntax. Print methods (both S3 and S4) are provided.  In order to
facilitate access to \code{animage} pixels methods have been defined
mirroring array access syntax (with \code{[]}) but taking into account
the value of \code{focus} and \code{outside}.}

\section{Slots}{
	 \describe{
    \item{\code{data}:}{Object of class \code{"array"} represents the
	  (possibly) non costant rectangular set of pixels}
    \item{\code{pmode}:}{Object of class \code{"character"} should match
	the string "scalar"}
    \item{\code{focus}:}{Object of class \code{"integer"} represents the 2D
	position of \code{data} in the infinite image region}
    \item{\code{outside}:}{Object of class \code{"numeric"} represents
	  the costant value assumed by the image outside the region
	  described by \code{data}}
    \item{\code{storage}:}{Object of class \code{"character"}  represents
	  the storage mode of \code{data} and can be \code{integer} or \code{real}}
    \item{\code{document}:}{Object of class \code{"logical"} should be
	  \code{TRUE} if we want to trigger the self documenting
	  capabilities of AnImAl on the specific object }
    \item{\code{history}:}{Object of class \code{"call"} representing
	  the processing history of the object. The object
	  \code{call("<undef>")} is used to identify a missing history.}
  }
}

\usage{
animage(data, pmode,
        focus = c(0L,0L), outside  = 0L, storage  = "integer",
        document = FALSE, history  = call("<undef>"))
as.animage(x)
is.animage(x)
is.animageList(x)
is.documentedAnimageList(x)
}

\arguments{
  \item{data}{see class slots}
  \item{pmode}{''}
  \item{focus}{''}
  \item{outside}{''}
  \item{storage}{''}
  \item{document}{''}
  \item{history}{''}
  \item{x}{an object}
}

\details{
}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
	\link{aniOperations}, \code{\linkS4class{animask}}
}
\examples{
#
# inspect the class
#
showClass("animage")
#
# create a small (3x3) image, activating self-documentation
#
img1 <- animage(array(c(1,2,3,
                        1,1,1,
                        8,8,8), c(3,3)),
                focus = c(1,1),
                document = TRUE)
#
# 'print' it
#
img1
#
# inspect its history
#
img1@history
#
# and regenerate it by evaluating its history
#
img1Clone <- eval(img1@history)
}
\keyword{classes}
