\name{ia.affine}

\alias{ia.affine}

\title{Image affine transformation}

\description{

   Apply an affine transformation to  an  integer (real) \code{animage}

}

\usage{ia.affine(img, affineMat)}

\arguments{
   	\item{img}{an integer (real) \code{animage}}
	\item{affineMat}{a 3x3, real, two dimensional matrix}
}

\value{
 An \code{animage}.
}

\details{
  
  The matrix \code{affineMat} specifies the affine transform in a non
  standard way:  it is a  3x3 array, whose  2x2 upper left  contains the
rotation/scale/shear part (R),  and the first two elements  of the third
column contain the  offset information (O): x' =  R(x+O) (instead of the
more  usual  x'=Rx+O).  The  third  row is  ignored.   Interpolation  is
performed using a Lanczos  kernel and preconditioning (smoothing) of the
image should  be performed  when downscaling it  (e.g. using  a Gaussian
kernel with sigma equal to sqrt(2)/2 for a factor 2 decimation).
  
}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
  \code{\link{ia.gaussianSmooth}}
}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages")
#
# Read a sample image
#
sampleFace <- as.animage(getChannels(read.pnm(file.path(sampleimages,
"sampleFace_01.pgm"))))
#
# smooth it
#
img <- ia.gaussianSmooth(sampleFace, sqrt(2)/2)
#
# rotation by 30 degrees, scale factor 0.5, and offset (20,40)
# combined
#
affineMat <- t(array(c(0.4330127, -0.2500000, 20,
                       0.2500000,  0.4330127, 40,
                               0,          0,  1),
                   dim=c(3,3)))
#
res <- ia.affine(img, affineMat)
#
# have a look
#
ia.show(res)
}
\keyword{array}
