\name{ia.cam}

\alias{ia.camNoiseRemovalLut}
\alias{ia.cam}

\title{Cellular automata machines support}

\description{

  Evolution of simple cellular automata machines with LUT based
  state transition specification

}

\usage{

ia.cam(img,  stateBits, stateLut, neighborhood = 4, steps = 1)
ia.camNoiseRemovalLut(neighborhood = 4)

}

\arguments{
  
   	\item{img}{an integer \code{animage}}
	\item{stateBits}{the number of bits necessary to represent a pixel state}
	\item{stateLut}{a list with \code{lutData} and \code{lutOffset} fields}
	\item{neighborhood}{must be 4 or 8, the number of neighboring pixels}
	\item{steps}{specifying the number of evolution time steps}

  }

\value{
  Function \code{ia.cam} returns an \code{animage}.

  Function    \code{ia.camNoiseRemovalLut}   returns    a    list   with
  \code{lutData} and \code{lutOffset} fields.

}

\details{
  
  The  function  \code{ia.cam} implements a simple cellular
  automata.   The bits  representing  the  state of  the  pixels in  the
neighborhood  of a  given pixel  (inclusive)  are packed  into a  single
integer number.  The  provided look up table is used  to map this number
into   the   new  value   for   the   central   pixel.  The   value   of
\code{neighborhood} specifies  the number of pixels  in the neighborhood
(not counting the central one). Function \code{ia.camNoiseRemovalLut}
create a look up table defining a denoising cellular automata machine
for binary images.


}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}
\references{
  T. Toffoli, N. Margolus, \emph{Cellular Automata Machines}, The MIT
  Press, 1987
}

\seealso{
  \code{\link{ia.table}}
}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages")
#
# Read a sample image
#
sampleFace <- as.animage(getChannels(read.pnm(file.path(sampleimages,
"sampleFace_01.pgm"))))
#
# add some noise
#
noisySampleFace <- ia.apply(sampleFace,
                            function(x,y,z) {z+rnorm(1, mean=0, sd=0.1)})
#
binaryNoisySampleFace <- ia.integerAnimage(ia.less(noisySampleFace, 0.5))
#
oldpar        <- par(no.readonly = TRUE)
#
par(mfrow   = c(2,2),
    mar     = c(3,0,0,0),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1.5)
#
plot(binaryNoisySampleFace)
#
for(i in 1:3) {
 binaryNoisySampleFace <- ia.cam(binaryNoisySampleFace,
                                 1L,
                                 ia.camNoiseRemovalLut(neighborhood=8),
                                 neighborhood=8,
                                 steps=2)
 plot(binaryNoisySampleFace)
}
#
mtext("A simple denoising cellular automata machine",
      outer = TRUE, side = 3, cex=1.5, line =1)
par(oldpar)
#
}
\keyword{array}
