\name{ia.convolution}

\alias{ia.convolution}

\title{Convolution with a generic kernel}

\description{

   Convolve an  integer (real) \code{animage} with an integer (real) kernel

}

\usage{ia.convolution(img, kernel)}

\arguments{
   	\item{img}{an integer (real) \code{animage}}
	\item{kernel}{a 2-dimensional integer (real) array with odd dimensions (1
	  is a valid dimension)}
}

\value{
 An \code{animage}.
}

\details{

The  function  computes  the convolution of the given \code{animage}
with the provided kernel. By providing suitable kernels, the function
can be used to perform a variety of tasks such as  smoothing, edge
detection, and sharpening.


}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages")
#
# Read a sample images
#
charA <- ia.readAnimageChannel(file.path(sampleimages, "a.tif"), 1)
#
# set the animage outside to white so that we do not get
# spurious effects at the boundary 
#
charA@outside <- 255L
#
# define a kernel to locate vertically oriented edges
#
kv    <- array(as.integer(c(0, 0,0,0,0,
                            0,-1,0,1,0,
                            0,-1,0,1,0,
                            0,-1,0,1,0,
                            0, 0,0,0,0)), c(5, 5))
#
# and one for horizontally oriented ones
#
kh    <- t(kv)
#
# compute the corresponding images
#
hedge <- ia.convolution(charA, kh)
vedge <- ia.convolution(charA, kv)
#
# have a look
#
ia.show(vedge)
#
# define a kernel for smoothing
#
ks    <- array(as.integer(c(1,1,1,1,1,
                            1,1,1,1,1,
                            1,1,1,1,1,
                            1,1,1,1,1,
                            1,1,1,1,1)), c(5, 5))
#
# and apply it 
#
smoothed  <- ia.div(ia.convolution(charA, ks), sum(ks))
#
# reset outside value 
#
smoothed@outside <- 255L
#
# and smooth more
#
smoothed2         <- ia.div(ia.convolution(smoothed, ks), sum(ks))
smoothed2@outside <- 255L
#
# have a look
#
ia.show(smoothed2)
#
}
\keyword{array}
