\name{ia.eps}

\alias{ia.eps}

\title{Print animage to eps file}

\description{

   Prints   an \code{animage} to an EPS file, providing controls for
   range of levels and monochrome output (by means of weighted channel average
   and gamma correction).

}

\usage{
ia.eps(img, file,
       srcRange = c(0,255), dstRange  =  c(0.05,0.95),
       stretch = FALSE, gamma = 1.0, color = FALSE,
       channels = c(1,2,3), weights = c(1,1,1))
}

\arguments{
   \item{img}{an animage object}
   \item{file}{a string representing the EPS filename (including the suffix)}
   \item{srcRange}{a vector of length 2 giving the black and white
	 levels for the original image}
   \item{dstRange}{a vector of length 2 giving the black and white
	 levels for the eps image}
   \item{channels}{a vector identifying image channels}
   \item{weights}{a vector identifying the weights attributed to \code{channels}}
   \item{stretch}{A boolean telling whether the minimum and maximum
	 value of the original image should be mapped to \code{dstRange}} 
   \item{gamma}{a real number specifying gamma correction}
   \item{color}{a boolean specifying whether the EPS image should be
	 color or monochrome}
}

\details{

  Monochrome images  are printed by default.  If  monochrome printing is
required,  a multichannel image  (provided as  a list  of \code{animage}
objects) needs  to be  transformed.  The default  transformation assumes
that  the input  image is  a three  channel one  and generates  a single
channel by simple average. The user can specify which channels should be
used and the corresponding weights to tune the results.  Additionally, a
gamma correction factor  $\gamma$ can be specified: when  $\gamma > 1.0$
you get a darker image, when $\gamma < 1.0$ you get a lighter image.  In
order to better  support the printing process, you  can also specify the
black and white levels using \code{normalize}: the first value specifies
that  the lowest  value in  the  original image  will be  mapped to  the
specified  fraction  of  a  fully  lit pixel,  while  the  second  value
specifies the mapping  for the image highlights. There  is currently non
support  within AnImAl  to  read an  EPS  images.  Due  to  the way  the
\code{pixmap}  package manages image  ranges, \code{dstRange}  should be
included within $[0,1]$ to avoid autostretching of the values.}


\value{
  The function returns \code{TRUE} upon successfull completion.
}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
   \link{aniUtilities},\link{aniIO}
}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages")
#
# Read the 'animage'
#
sampleFace   <- ia.readAnimage(file.path(sampleimages, "sampleFace_06.tif"))
#
# and  create a 'lighter' monochrome EPS file using only the 'blue' channel
#
success <- ia.eps(sampleFace, "sampleFace.eps", 
                  srcRange = c(0,255), dstRange = c(0.05,0.5),
                  stretch = TRUE, gamma = 1/2.0,
                  channels = c(1,2,3), weights = c(0,0,1))
#
}
\keyword{file}
