\name{ia.gradient}

\alias{ia.gradient}

\title{Gradient computation for grey level images}

\description{

   Compute the horizontal and vertical derivatives of a grey level
   image, and, optionally, the corresponding gradient direction

}

\usage{ia.gradient(img, Kx = array(c(-1,0,1), c(1,3)),
                        Ky,
                        smoothing = 1.0,
                        metrics   = 1,
                        direction = FALSE,
                        thr       = 1e-15)}

\arguments{
   	\item{img}{an integer (real) \code{animage}}
	\item{Kx}{a 2-dimensional real array with odd dimensions (1 is a valid dimension) to compute the x directional   derivative}
    \item{Ky}{the same as \code{Kx} but for the y (vertical) directional derivative}
    \item{smoothing}{a real value. If positive, it determines the sigma
of the regularizing Gaussian kernel.}
\item{metrics}{1 or 2:specifies the metrics used to compute gradient magnitude.}
    \item{direction}{a logical: if \code{TRUE} compute gradient
direction as well.}
    \item{thr}{a real: direction when gradient magnitude is below
threshold is set to zero.}
}

\value{
 An \code{animage} or a list of \code{animage}s representing the
magnitude or the magnitude and direction of the gradient.
}

\details{

The function computes the gradient magnitude by convolving the image
with the supplied kernels to obtain the directional derivatives along x
and y. If the kernel for y is not specified the transpose of \code{Kx}
is used. In order to get a better estimate, especially when directions
are required, a positive \code{smoothing} factor can be
specified. Threshold \code{thr} is used to avoid noisy directional
estimates when the gradient magnitude is very small.

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages")
#
# Read a sample image
#
square <- as.animage(getChannels(read.pnm(file.path(sampleimages,
"square.pgm"))))
square@outside <- 1
#
oldpar        <- par(no.readonly = TRUE)
par(mfrow   = c(2,2),
    mar     = c(2,1,1,1),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1)
#
ia.show(square, xlab="Original")
ia.show(ia.gradient(square, smoothing=0), xlab="Gradient magnitude")
info <- ia.gradient(square, smoothing=4, direction=TRUE)
ia.show(info$magnitude, xlab = "Magnitude (smoothing)")
ia.show(info$direction, xlab = "Direction (smoothing)")
#
par(oldpar)
#
}
\keyword{array}
