/*
 * (c) 1993,1994,1995,1996 by ITC-IRST
 * 
 * Latest update 24/06/2003
 */



#ifndef	_array_h
#define _array_h 1

#include	<cstdlib>
#include	<cmath>
#include	<iostream>



#include <arrayTools/namespace_header.h>
#include <arrayTools/global.h>
#include <arrayTools/Vector.h>
#include <arrayTools/io_heads.h>
#include <arrayTools/Endian.h>
#include <arrayTools/ArrayIter.h>
#include <arrayTools/Linear_base.h>


#define SIZE 128000

namespace _ARRAYTOOLS_NAMESPACE
{

  extern int	created_arrays;	//! Count the arrays created within the class
  extern int	deleted_arrays;	//! Count the arrays destroyed within the class
  extern void     ArrayError();
  extern void     ArrayError(char *message);

  void ArrayCount(char* label);

  static vector<int> foo;

  template<class T> class array 
    {
      vector<T>	**elts;
      T	        **quick;
      char	reuse;
      int	first_row_idx;
      int	first_col_idx;
      int	last_row_idx;
      int	last_col_idx;
      int	nr_of_rows;
      int	nr_of_cols;
      float     scaling;

    private:
      void		FreeOwnStorage();
      void		TransferStorage(int	        *ret_first_row_idx,
					int	        *ret_first_col_idx,
					int	        *ret_nr_of_rows,
					int 	        *ret_nr_of_cols,
					int	        *ret_last_row_idx,
					int	        *ret_last_col_idx,
					vector<T>	***ret_elts,
					T	        ***ret_quick);
    
    public:
      array();
      array(unsigned int, unsigned int);
      array(int, int, int, int);
      array(array<T> &); 
      array(const array<T> &);
      array(vector<T> &);
      ~array();



      typedef _array_iterator<T, T&, T*> iterator;
      typedef _array_iterator<T, const T&, T*> const_iterator;


      iterator begin()
	{
	  return iterator(this, first_row_idx, first_col_idx);
	}

      iterator end()
	{
	  return ( is_empty() ? 
		   iterator(this, 0, 0) :
		   iterator(this, last_row_idx+1, 0) );
	}




      // handle temporary obj
      inline int	dispose() const { return((int) reuse);}
      array<T>&		temporary();


      array<T>&		translate_at(int, int);
      array<T>&         at(array<T>&);
      array<T>&		subimage(int, int, int, int);
   
    
      // content access
      T**	                storage();
      inline  vector<T>&	operator[](int);
    
    
      inline void       stat(int& x0, int& x1, int& y0, int& y1);
      inline void       boundary(int &y0, int &y1, int &x, int &x1);

      int		rows() const     	 { return(nr_of_rows);    }
      int		cols() const    	 { return(nr_of_cols);    }
      int	        height() const    	 { return(nr_of_rows);    }
      int		width() const    	 { return(nr_of_cols);    }
      int		first_row_index() const  { return(first_row_idx); }
      int		last_row_index() const   { return(last_row_idx);  }
      int		first_col_index() const  { return(first_col_idx); }
      int		last_col_index() const   { return(last_col_idx);  }
      void              scale(float f)           { scaling = f;}


      //reset: delete and reallocate storage
      void                          reset();
      void                          reset(int rows, int cols);
      void                          reset(int r0, int r1, int c0, int c1);
      template<typename U> void        reset(const array<U> & A);
      
      //resize: redim storage if necessary
      //return TRUE if memory is reallocated
      int                           resize(int rows, int cols);
      int                           resize(int r0, int r1, int c0, int c1);
      template<typename U> int      resize(const array<U> & A);

      int                           is_empty();

      //operators

      array<T>&		operator=(const array<T>&);
      array<T>&		operator=(array<T>&);
      array<T>&		operator=(double);
      array<T>&		operator-();
      array<T>&		operator+(array<T>&);
      array<T>&		operator-(array<T>&);
      array<T>&		operator*(array<T>&);
      array<T>&		operator/(array<T>&);
      array<T>&		operator&(array<T>&);
      array<T>&		operator%(array<T>&);
      array<T>&		operator>(array<T>&);
      array<T>&		operator<(array<T>&);
      array<T>&		operator>(double);
      array<T>&		operator<(double);
      array<T>&		operator==(double);
      array<T>&		operator+=(double);
      array<T>&		operator*=(double);
      array<T>&		operator!();
      array<T>&		operator&&(array<T>&);
      array<T>&		operator||(array<T>&);
      array<T>&	        operator+(double);
      array<T>&	        operator-(double);
      array<T>&	        operator*(double);
      array<T>&	        operator/(double);
      array<T>&	        operator^(int);


      boolean	        operator==(array<T>&);
      boolean	        operator!=(array<T>&);

      array<T>&             map(T (*f)(T a));

      //methods

      void                  write(ostream&);
      void                  write(ostream&, int);
      void                  write(char *filename);
      void                  write(char *filename, int fmt);
      void                  writeAscii(ostream&);
      void                  writeAscii(ostream&, int);
      void                  writeAscii();
      void                  writeAscii(char*,    int, char*);
      void                  readAscii(char*);
      void                  readAscii(char*, int i0, int i1);
      void                  readAscii(char*, vector<int> &I);
      void                  read(char*);
      void                  read(istream&);
      void                  read(istream&, int i0, int i1);
      void                  read(istream&, vector<int> &I);
      void                  readAsSingleRow(istream&);
      void                  readAsSingleRow(istream&, int i0, int i1);
      void                  readAsSingleRow(istream&, vector<int> &I);
      void                  get(char*);
      void                  get(char*, int i0, int i1);
      void                  get(char*, vector<int> &I);

    };



  
  template<class T> inline void	DeleteIfTemporary(array<T> &);
  template<class T, class U> inline void CheckArraysConsistency(const array<T> &A, const array<U> &B);

  template<class T, class U> array<T> & Cast(array<U>& B, array<T>& A);
  template<class T, class U> array<T> & CastTo(array<U> & A);

  template<class T> array<T>&	        abs(array<T>& A);
  template<class T> array<T>&	        log(array<T>& A);
  template<class T> array<T>&	        sqrt(array<T>& A);
  template<class T> array<T>&	        Tr(array<T>& A);
  template<class T> double 	        Det(array<T>& A);
  template<class T> array<T>&	        Inverse(array<T>& A);

  template<class T> array<T>& GaussianSmooth(array<T>& A, double sigma);
  template<class T> array<T>& GaussianSmooth2(array<T>& A, double sigma);
  template<class T> array<T>& Unzoom        (array<T>& A, int rx, int ry);
  template<class T> array<T>& Zoom          (array<T>& A, int rx, int ry);
  template<class T> array<T>& Grow          (array<T>& A, int n);
  template<class T> array<T>& Shrink        (array<T>& A, int n);
  template<class T> array<T>& Grow          (array<T>& A, int m, int n);
  template<class T> array<T>& Shrink        (array<T>& A, int m, int n);
  template<class T> array<T>& Linearize(array<T>& A);
  

  //==================================================================================================
  // METHODS

  
 

  //! Free own data storage, and reset internal  data to an 
  //! empty stab
  //!

  template<class T> void array<T>::FreeOwnStorage()
    {
      int	i;
   
      if(elts + first_row_idx != NULL)
	{
	  for(i = first_row_idx; i <= last_row_idx; i++)
	    delete elts[i];

	  elts	+= first_row_idx;
	  quick	+= first_row_idx;
	  delete[] elts;
	  delete[] quick;
	}

      elts 		= NULL;
      quick		= NULL;
      first_row_idx   	= 0;
      first_col_idx   	= 0;
      last_row_idx         = -1;
      last_col_idx         = -1;
      nr_of_rows		= 0;
      nr_of_cols		= 0;
    }
 

  //! WARNING: this function must be used with extrem care.
  //! It returns via a set of pointers all its internal 
  //! structure, and reset itself to an empty stab.
  //! It is used to improve assignment efficiency,
  //! by transferring data of a volatile array to the 
  //! LHS of an assignment statement.
  //!

  template<class T> void array<T>::TransferStorage(int       *ret_first_row_idx,
						   int       *ret_first_col_idx,
						   int       *ret_nr_of_rows,
						   int       *ret_nr_of_cols,
						   int       *ret_last_row_idx,
						   int       *ret_last_col_idx,
						   vector<T> ***ret_elts,
						   T	   ***ret_quick)
    {
      *ret_elts 		= elts;
      *ret_quick 		= quick;
      *ret_first_row_idx	= first_row_idx;
      *ret_first_col_idx	= first_col_idx;
      *ret_nr_of_rows	= nr_of_rows;
      *ret_nr_of_cols	= nr_of_cols;
      *ret_last_row_idx	= last_row_idx;
      *ret_last_col_idx	= last_col_idx;

      elts 		= NULL;
      quick 		= NULL;
   
      first_row_idx   	= 0;
      first_col_idx   	= 0;
      last_row_idx         = -1;
      last_col_idx         = -1;
      nr_of_rows		= 0;
      nr_of_cols		= 0;
    }
    


  //! Create an array with no associated elements 
  //!

  template<class T> array<T>::array()
    {
      created_arrays++;

      elts			= NULL;
      quick		= NULL;
      reuse		= FALSE;
      first_row_idx   	= 0;
      first_col_idx   	= 0;
      last_row_idx		= -1;
      last_col_idx		= -1;
      nr_of_rows		= 0;
      nr_of_cols		= 0;
      scaling              = 1.0;
    }


  //! Create an array with <rows> rows and <cols> columns
  //! 	  indexes for both rows and columns start at zero
  //! 	  Standard C-language syntax
  //!

  template<class T> array<T>::array(unsigned int rows,	
				    unsigned int cols)	
    {
      int		i;
   
      elts			= new vector<T> *[rows];
      quick		= new T         *[rows];
      reuse		= FALSE;
      first_row_idx   	= 0;
      first_col_idx   	= 0;
      last_row_idx        	= rows - 1;
      last_col_idx        	= cols - 1;
      nr_of_rows		= rows;
      nr_of_cols		= cols;
      scaling              = 1.0;

      for(i = first_row_idx; i <= last_row_idx; i++)
	{
	  elts[i] 	= new vector<T>(nr_of_cols);
	  quick[i]	= (*elts[i]).storage();
	}

      created_arrays++;
    }


  //! Create an array with <lr_idx - fr_idx + 1> rows and 
  //!        <lc_idx - fc_idx + 1> columns
  //! 	  indexes for rows start at fr_idx
  //!	  indexes for columns start at fc_idx
  //!

  template<class T> array<T>::array(int fr_idx, int lr_idx,	
				    int fc_idx, int lc_idx)	
    {
      int		i;
  
      if((fr_idx > lr_idx) ||  (fc_idx > lc_idx))
	{
	  cerr << "Illegal range for array indexes: negative size implied" << endl;
	  cerr << fr_idx << " " << lr_idx << " " << fc_idx << " " << lc_idx << endl;
	  exit(-1);
	}

      reuse		= FALSE;
      first_row_idx   	= fr_idx;
      first_col_idx   	= fc_idx;
      last_row_idx        	= lr_idx;
      last_col_idx        	= lc_idx;
      nr_of_rows		= lr_idx - fr_idx + 1;
      nr_of_cols		= lc_idx - fc_idx + 1;
      scaling              = 1.0;
      elts			= new vector<T> *[nr_of_rows];
      quick		= new T         *[nr_of_rows];
   
      elts			-= first_row_idx;
      quick		-= first_row_idx;
   
      for(i = first_row_idx; i <= last_row_idx; i++)
	{
	  elts[i] 	= new vector<T>(first_col_idx, last_col_idx);
	  quick[i]	= (*elts[i]).storage();
	}

      created_arrays++;
    }


  //! Create a copy of an existing array. 
  //!

  template<class T> array<T>::array(array<T>& A)
    { 
      //cerr << "array<T>::array(array<T>& A)\n";
      
      unsigned int	r = A.rows();
      unsigned int	c = A.cols();
      int		i;

   
      first_row_idx	= A.first_row_index();
      first_col_idx	= A.first_col_index();
      last_row_idx		= A.last_row_index();
      last_col_idx		= A.last_col_index();
      scaling              = 1.0;
      
      if((r > 0) && (c > 0))
	{
	  elts	= new vector<T> *[r];
	  quick	= new T         *[r];
	
	  elts 	-= first_row_idx;
	  quick 	-= first_row_idx;
   
	  for(i = first_row_idx; i <= last_row_idx; i++)
	    {
	      elts[i] 	= new vector<T>(A[i]);
	      quick[i]	= (*elts[i]).storage();	
	    }
	
	}
      else if ((r  == 0) && (c == 0))
	{
	  elts	= NULL;
	  quick	= NULL;
	}
      else
	{
	  cerr
	    << "Illegal array dimensions ["
	    << r
	    << ", "
	    << c
	    << "]\n";
	  exit(-1);
	}
   
   
      reuse	= FALSE;
      nr_of_rows	= r;
      nr_of_cols	= c;

      DeleteIfTemporary(A);

      created_arrays++;
    }

   
  //! Create a copy of a const existing array. 
  //!

  template<class T> array<T>::array(const array<T>& A)
    {
      // cerr << "array<T>::array(const array<T>& A)\n";

      unsigned int	r = A.rows();
      unsigned int	c = A.cols();
      int		i;

   
      first_row_idx	= A.first_row_index();
      first_col_idx	= A.first_col_index();
      last_row_idx	= A.last_row_index();
      last_col_idx	= A.last_col_index();
      scaling           = 1.0;
      
      if((r > 0) && (c > 0))
	{
	  T * destv;
	  T * srcv;

	  elts	= new vector<T> *[r];
	  quick	= new T         *[r];
	
	  elts 	-= first_row_idx;
	  quick -= first_row_idx;
   
	  for(i = first_row_idx; i <= last_row_idx; i++)
	    {
	      elts[i]  = new vector<T>(first_col_idx, last_col_idx);
	      destv    = elts[i]->storage();	
	      quick[i] = destv;
	      destv   += first_col_idx;
	      srcv     = A.elts[i]->storage() + first_col_idx;
	      memcpy(destv,srcv,c*sizeof(T));
	    }
	
	}
      else if ((r  == 0) && (c == 0))
	{
	  elts	= NULL;
	  quick	= NULL;
	}
      else
	{
	  cerr
	    << "Illegal array dimensions ["
	    << r
	    << ", "
	    << c
	    << "]\n";
	  exit(-1);
	}
   
   
      reuse	= FALSE;
      nr_of_rows	= r;
      nr_of_cols	= c;

      created_arrays++;
    }
  

  //! Transform a vector into an array
  //!

  template<class T> array<T>::array(vector<T>& v)
    {
      unsigned int	r = 1;
      unsigned int	c = v.length();

   
      first_row_idx	= v.first_index();
      first_col_idx	= v.first_index();
      last_row_idx		= first_row_idx;
      last_col_idx		= v.last_index();

      if((r > 0) && (c > 0))
	{
	  elts	= new vector<T> *[r];
	  quick	= new T         *[r];
	
	  elts 	-= first_row_idx;
	  quick 	-= first_row_idx;
   
	  elts[first_row_idx] = new vector<T>(v);
	  quick[first_row_idx]= (*elts[first_row_idx]).storage();
	}
      else if ((r  == 0) && (c == 0))
	{
	  elts	= NULL;
	  quick 	= NULL;
	}
      else
	{
	  cerr
	    << "Illegal array dimensions ["
	    << r
	    << ", "
	    << c
	    << "]\n";
	  exit(-1);
	}
   
   
      reuse	= TRUE;
      nr_of_rows	= r;
      nr_of_cols	= c;
      scaling      = 1.0;

      DeleteIfTemporary(v);

      created_arrays++;
    }


  //! Destroy an array by freeing each constituent vector 
  //! 	   and the vector of vector pointers
  //!

  template<class T>  array<T>::~array()
    {
      int	i;
   
      if(nr_of_rows == -1)
	{
	  cerr << "Internal error: trying to delete already deleted array!\n";
	  cerr << "                Improper use of temporary array?       \n";

	  ArrayError();
	
	  exit(-1);
	}

      if(elts + first_row_idx != NULL)
	{
	  for(i = first_row_idx; i <= last_row_idx; i++)
	    delete elts[i];

	  elts	+= first_row_idx;
	  quick	+= first_row_idx;
	  delete[] elts;
	  delete[] quick;
	}

      elts		= NULL;
      quick	= NULL;
      nr_of_rows	= -1;
   
      deleted_arrays ++;
    }

  template<class T> void  array<T>::reset()
    { 
      (*this) = (new array<T>)->temporary(); 
    }

  template<class T> void  array<T>::reset(int rows, int cols)  
    { 
      (*this) = (new array<T>(rows,cols))->temporary(); 
    }
  
  template<class T> void  array<T>::reset(int r0, int r1, int c0, int c1) 
    { 
      (*this) = (new array<T>(r0,r1,c0,c1))->temporary(); 
    }

  template<class T> template<typename U> void array<T>::reset(const array<U> & A)
    { 
      (*this).reset(A.first_row_index(), A.last_row_index(), 
		    A.first_col_index(), A.last_col_index()); 
    }

  template<class T> int array<T>::is_empty()
    { 
      if ( nr_of_rows == 0 || nr_of_cols == 0 ) return TRUE; 
      else return FALSE;
    }

  template<class T> int array<T>::resize(int rows, int cols)
    {
      if( nr_of_rows == rows && nr_of_rows == cols) return FALSE;
      (*this).reset(first_row_idx, first_row_idx+rows-1, first_col_idx, first_col_idx+cols-1);
      return TRUE;
    }

  template<class T> int array<T>::resize(int r0, int r1, int c0, int c1)
    {
      if(r1-r0+1 == nr_of_rows && c1-c0+1 == nr_of_cols)
	{
	  if(r0 == first_row_idx && c0 == first_col_idx)
	    {
	      return FALSE;
	    }
	  else
	    {
	      (*this).translate_at(r0,c0);
	      return TRUE;
	    }
	}
      else
	{
	  (*this).reset(r0,r1,c0,c1);
	  return TRUE;
	}
    }

  template<class T> template<typename U> int array<T>::resize(const array<U> & A)
    { 
      return (*this).resize(A.first_row_index(), A.last_row_index(), 
			   A.first_col_index(), A.last_col_index()); 
    }

  //! Reference operator for accessing the array rows
  //! If the package is compiled with the flag CHECKS undefined, 
  //! no checking is done on the access index.
  //!

  template<class T> vector<T>&	array<T>::operator[](int i)
    {
#ifndef CHECKS
      return(*elts[i]);
#else
      if(nr_of_rows == -1)
	{
	  cerr << "Illegal reference to elements of a deleted array!\n";
	  cerr << "        Improper use of temporary vector?       \n";
	}
      else if((i < first_row_idx) || (i > last_row_idx))
	{
	  char err_msg[256];
	     
	  sprintf(err_msg, "Array: row index out of range: %d not in [%d,%d]",
		  i, first_row_idx, last_row_idx);
	     
	  ArrayError(err_msg);

	  exit(-1);
	}

      return(*(elts[i]));
#endif   
    }


  //! Return a sub region of the given array as a new array tagged for deletion.
  //! In fact, the region can be greater than the original array, the extra
  //! elements  being set to zero.
  //!

  template<class T>  array<T>&	array<T>::subimage(int r0, int r1, int c0, int c1)
    {
      array<T>	*X;
      int		x, y;
      int		x0, y0;
      int		x1, y1;
      T            **_X;
      T            **_t;
   
      X	= new array<T>(r0, r1, c0, c1);

      y0	= MAX(r0, first_row_idx);
      x0	= MAX(c0, first_col_idx);

      y1	= MIN(r1, last_row_idx);
      x1	= MIN(c1, last_col_idx);

      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    _X[y][x] = _t[y][x];
	}


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Change base index for both rows and columns
  //!

  template<class T>  array<T>& array<T>::translate_at(int new_first_row_index, 
						      int new_first_col_index)
    {
      if(elts + first_row_idx != NULL)
	{
	  int	i;
	
	  elts		= elts  + (first_row_idx - new_first_row_index);
	  quick		= quick + (first_row_idx - new_first_row_index);
	  first_row_idx	= new_first_row_index;
	  first_col_idx	= new_first_col_index;
	  last_row_idx	= first_row_idx + nr_of_rows - 1;
	  last_col_idx	= first_col_idx + nr_of_cols - 1;

	  for(i = first_row_idx; i <= last_row_idx; i++)
	    {
	      (*elts[i]).translate_at(new_first_col_index);
	      quick[i]	= (*elts[i]).storage();
	    }
	}

      return(*this);
    }


  //! Modify an array by making it temporary
  //!

  template<class T>  array<T>& array<T>::temporary()
    {
      reuse = TRUE;
   
      return(*this);
    }


  //! Return the array values when the corresponding element
  //! of filter array A is greater than 0, else otherwise
  //!

  template<class T> array<T>& array<T>::at(array<T> &A)
    {
      int		x, y;
      array<T>	*X;
      T            **_X, **_A, **_t;
   
      CheckArraysConsistency(*this, A);
   
      X    = new array<T>(first_row_idx, last_row_idx,
			  first_col_idx, last_col_idx);

      _X   = (*X).storage();
      _t   = (*this).storage();
      _A   = A.storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (_A[y][x] > 0) ? _t[y][x] : 0;

      DeleteIfTemporary(A);
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }

  //return bounds
  template<class T> void	array<T>::stat(int &x0,
					       int &x1,
					       int &y0,
					       int &y1)
    {
      y0 = first_row_idx;
      y1 = last_row_idx;
      x0 = first_col_idx;
      x1 = last_col_idx;
    }

  //! Return (through referenced arguments) the boundary points
  //!

  template<class T> void	array<T>::boundary(int &y0,
						   int &y1,
						   int &x0,
						   int &x1)
    {
      y0 = first_row_idx;
      y1 = last_row_idx;
      x0 = first_col_idx;
      x1 = last_col_idx;
    }
    
  //! Update internal pointer structure for quick referencing
  //! and return the pointer to the internal storage
  //!

  template<class T> T**	array<T>::storage()
    {
      int	i;

      if(elts + first_row_idx != NULL)
	{
	  for(i = first_row_idx; i <= last_row_idx; i++)
	    quick[i] = (*elts[i]).storage();
	}
   
      return(quick);
    }
    


  //! Assignment operator.  Data of right hand side are copied into 
  //! the data region of the left hand side.
  //! If the RHS is tagged for deletion, it is deleted.
  //! The LHS is protected from deletion
  //!

  template<class T> array<T>&	array<T>::operator=(array<T>& A)
    {
      int	        r = A.rows();
      int	        c = A.cols();
      int		ro = A.first_row_index();
      int		co = A.first_col_index();
      int		i;
   
      if(A.dispose() == TRUE)
	{
	  //! If A is volatile, exploit its data
	  //! structure

	  (*this).FreeOwnStorage();
	  A.TransferStorage(&first_row_idx,
			    &first_col_idx,
			    &nr_of_rows,
			    &nr_of_cols,
			    &last_row_idx,
			    &last_col_idx,
			    &elts,
			    &quick);

	  reuse	= FALSE;
	  delete &A;
	}
      else
	{
	  if((r > 0) && (c > 0))
	    {
	      //! If the RHS is consistent with the LHS	
	      //!  do not allocate new vectors		
	     
	      if((r 	== nr_of_rows) 		&&
		 (c 	== nr_of_cols)		&&
		 (co 	== first_col_idx)		&&
		 (ro 	== first_row_idx))
		{
		  for(i = first_row_idx; i <= last_row_idx; i++)
		    {
		      *(elts[i]) 	= A[i];
		      quick[i] 	= (*elts[i]).storage();
		    }
		}
	      else
		{
		  if(elts + first_row_idx != NULL)	
		    {
		      for(i = first_row_idx ; i <= last_row_idx; i++)
			delete elts[i];
		       
		      elts	+= first_row_idx;
		      quick	+= first_row_idx;
		       
		      delete[] elts;
		      delete[] quick;
		    }
		  
		  elts		= new vector<T> *[r];
		  quick		= new T         *[r];
		  
		  first_row_idx	= A.first_row_index();
		  first_col_idx	= A.first_col_index();
		  last_row_idx	= A.last_row_index();
		  last_col_idx	= A.last_col_index();
		  
		  elts		-= first_row_idx;
		  quick		-= first_row_idx;
		  
		  for(i = first_row_idx; i <= last_row_idx; i++)
		    {
		      elts[i]	= new vector<T>(A[i]);
		      quick[i]	= (*elts[i]).storage();
		    }
		}
	    }
	  else if ((r == 0) && (c == 0))
	    {
	      if(elts + first_row_idx != NULL)	
		{
		  for(i = first_row_idx ; i <= last_row_idx; i++)
		    delete elts[i];
		  
		  elts	+= first_row_idx;
		  quick	+= first_row_idx;
		  
		  delete[] elts;
		  delete[] quick;
		}
	     
	      elts		= NULL;
	      quick 		= NULL;
	     
	      first_row_idx	= A.first_row_index();
	      first_col_idx	= A.first_col_index();
	      last_row_idx	= A.last_row_index();
	      last_col_idx	= A.last_col_index();
	     
	    }
	  else
	    {
	      cerr << "Inconsistent RHS for assignment\n";
	      exit(-1);
	    }
	
	
	  reuse		= FALSE;
	  nr_of_rows	= r;
	  nr_of_cols	= c;
	
	  DeleteIfTemporary(A);
	}
   
      return (*this);
    }

  //! Assignment operator with const argument
  //!

  template<class T> array<T>& array<T>::operator=(const array<T>& A)
  {
    //cerr << "array<T>::operator=(const array<T>& A)\n";

    int	        r = A.rows();
    int	        c = A.cols();
    int		ro = A.first_row_index();
    int		co = A.first_col_index();
    int		i;

    if(A.dispose() == TRUE)
      {
	fprintf(stderr,"array<T>::operator=(const array<T>&) : warning, RHS dispose true\n");
      }

    if((r > 0) && (c > 0))
      {
	//! If the RHS is consistent with the LHS	
	//!  do not allocate new vectors		
	     
	if((r 	== nr_of_rows) 		&&
	   (c 	== nr_of_cols)		&&
	   (co 	== first_col_idx)	&&
	   (ro 	== first_row_idx))
	  {
	    for(i = first_row_idx; i <= last_row_idx; i++)
	      {
		*(elts[i]) 	= *(A.elts[i]);
		quick[i] 	= (*elts[i]).storage();
	      }
	  }
	else
	  {
	    if(elts + first_row_idx != NULL)	
	      {
		for(i = first_row_idx ; i <= last_row_idx; i++)
		  delete elts[i];
		       
		elts	+= first_row_idx;
		quick	+= first_row_idx;
		       
		delete[] elts;
		delete[] quick;
	      }
		  
	    elts		= new vector<T> *[r];
	    quick		= new T         *[r];
		  
	    first_row_idx	= A.first_row_index();
	    first_col_idx	= A.first_col_index();
	    last_row_idx	= A.last_row_index();
	    last_col_idx	= A.last_col_index();
		  
	    elts		-= first_row_idx;
	    quick		-= first_row_idx;
		  
	    for(i = first_row_idx; i <= last_row_idx; i++)
	      {
		elts[i]	= new vector<T>(*(A.elts[i]));
		quick[i]	= (*elts[i]).storage();
	      }
	  }
      }
    else if ((r == 0) && (c == 0))
      {
	if(elts + first_row_idx != NULL)	
	  {
	    for(i = first_row_idx ; i <= last_row_idx; i++)
	      delete elts[i];
		  
	    elts	+= first_row_idx;
	    quick	+= first_row_idx;
		  
	    delete[] elts;
	    delete[] quick;
	  }
	     
	elts		= NULL;
	quick 		= NULL;
	     
	first_row_idx	= A.first_row_index();
	first_col_idx	= A.first_col_index();
	last_row_idx	= A.last_row_index();
	last_col_idx	= A.last_col_index();
	     
      }
    else
      {
	cerr << "Inconsistent RHS for assignment\n";
	exit(-1);
      }
	
	
    reuse		= FALSE;
    nr_of_rows	= r;
    nr_of_cols	= c;
   
    return (*this);
  } 
  


  //! Substitute array value vith the given value
  //!

  template<class T>  array<T>&	array<T>::operator=(double value)
    {
      int  x, y;
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  (*this)[y][x] = T(value);
   
      return (*this);    
    }



  //! Return TRUE if the two arrays are equal.
  //!

  template<class T> boolean	array<T>::operator==(array<T>& A)
    {
      int		ro	= A.first_row_index();
      int		co	= A.first_col_index();
      int		rl	= A.last_row_index();
      int		r	= A.rows();
      int		c	= A.cols();
      int		i;
      boolean	b 	= TRUE;
   
      if((ro != first_row_idx)	||
	 (co != first_col_idx)	||
	 (r  != nr_of_rows)			||
	 (c  != nr_of_cols))
	{
	  b	= FALSE;
	}
      else
	{
	  for(i = ro; (i <= rl) && (b == TRUE); i++)
	    b	&= (A[i] == *elts[i]) ? TRUE : FALSE;
	}
   
      DeleteIfTemporary(*this);
      DeleteIfTemporary(A);
   
      return(b);
    }


  //! Return TRUE if two arrays differ
  //!

  template<class T> boolean	array<T>::operator!=(array<T>& A)
    {
      return((*this == A) ? FALSE : TRUE);
    }


  //! Add array, element by element.
  //! If the arrays have different dimensions or base indexes an error 
  //! occurs.
  //! The result of the operation is tagged for deletion after usage
  //!

  template<class T> array<T>&	array<T>::operator+(array<T> &A)
    {
      array<T>	*X;
      T            **_X, **_t;
      int          x, y;

      CheckArraysConsistency(*this, A);
   
      X    = new array<T>(A);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = _X[y][x] + _t[y][x];
   
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Unary minus.
  //! The result of the operation is tagged for deletion after usage
  //!

  template<class T> array<T>&	array<T>::operator-()
    {
      int           x, y;
      array<T>	 *X;
      T             **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx,
		       first_col_idx, last_col_idx);
   
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = -_t[y][x];
	 
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }



  //! Subtract array, element by element.
  //! If the arrays have different dimensions or base indexes an error 
  //! occurs.
  //! The result of the operation is tagged for deletion after usage
  //!

  template<class T> array<T>&	array<T>::operator-(array<T> &A)
    {
      array<T>	*X;
      T            **_X, **_A, **_t;
      int          x, y;
   
      CheckArraysConsistency(*this, A);
   
      X    = new array<T>(first_row_idx, last_row_idx, 
			  first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _A   = A.storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = _t[y][x] - _A[y][x];
  
      DeleteIfTemporary(A);
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Multiply array, element by element.
  //! If the arrays have different dimensions or base indexes an error 
  //! occurs.
  //! The result of the operation is tagged for deletion after usage
  //!

  template<class T> array<T>&	array<T>::operator&(array<T> &A)
    {
      array<T>	*X;
      T            **_X, **_t;
      int          x, y;
      
      CheckArraysConsistency(*this, A);
   
      X    = new array<T>(A);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = _t[y][x] * _X[y][x];
   
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Divide array, element by element.
  //! If the arrays have different dimensions or base indexes an error 
  //! occurs.
  //! The result of the operation is tagged for deletion after usage
  //!

  template<class T> array<T>&	array<T>::operator%(array<T> &A)
    {
      array<T>	*X;
      T            **_X, **_t;
      int          x, y;
      
      CheckArraysConsistency(*this, A);
   
      X    = new array<T>(A);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = _t[y][x] / _X[y][x];
   
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Apply to each array element the given function
  //!

  template<class T> array<T>& array<T>::map(T (*f)(T a))
    {
      array<T>	*X;
      T            **_X;
      int          x, y;
      
      X    = new array<T>(*this);
      _X   = (*X).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (*f)(_X[y][x]);
   
      return((*X).temporary());
   
    }
    

  //! Row-column multiply.
  //! If the arrays have different  base indexes an error occurs.
  //! The result of the operation is tagged for deletion after usage.
  //!

  template<class T> array<T>&	array<T>::operator*(array<T> &A)
    {
      array<T>	At = Tr(A);
      int		x, x0, x1;
      int		y, y0, y1;
      array<T>	*X;
   
      if((first_col_idx  != At.first_col_index())	||
	 (nr_of_cols     != At.cols()))
	{
	  cerr << "Cannot multiply arrays: unmatched base indexes or dimensions!\n";
	  exit(-1);
	}
   
      y0	= first_row_idx;
      y1	= last_row_idx;
      x0	= At.first_row_index();
      x1	= At.last_row_index();
   
      X	= new array<T>(y0, y1, x0, x1);
   
      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    (*X)[y][x] = T(*elts[y]  * At[x]);
	}
   
      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Multiply by inverse of <A>.
  //! Result is tagged for deletion.
  //!

  template<class T> array<T>& array<T>::operator/(array<T>& A)
    {
      return((*this)*Inverse(A));
    }

  //! Apply to each element the standard arithmetic operators with a scalar argument
  //! operators
  //!

  template<class T> array<T>&		array<T>::operator+(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (T)(_t[y][x] + v);


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }

  template<class T> array<T>&		array<T>::operator-(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = _t[y][x] - v;


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }

  template<class T> array<T>&		array<T>::operator*(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (T)(_t[y][x] * v);


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }

  template<class T> array<T>&		array<T>::operator/(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (T)(_t[y][x] / v);


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  //! Apply to each element the standard relational 
  //! operators
  //!

  template<class T> array<T>&		array<T>::operator>(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (_t[y][x] > v) ? 1 : 0;


      DeleteIfTemporary(*this);
   
      return((*X).temporary());
    }


  template<class T> array<T>&		array<T>::operator<(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (_t[y][x] < v) ? 1 : 0;

      DeleteIfTemporary(*this);

      return((*X).temporary());
    }

  template<class T> array<T>&		array<T>::operator==(double v)
    {
      int		y, x;
      array<T>	*X;
      T            **_X, **_t;
   
      X	= new array<T>(first_row_idx, last_row_idx, 
		       first_col_idx, last_col_idx);
      _X   = (*X).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (_t[y][x]== v) ? 1 : 0;

      DeleteIfTemporary(*this);

      return((*X).temporary());
    }

  template<class T> array<T>&		array<T>::operator+=(double v)
    {
      int		y, x;
      T            **_t;
   
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _t[y][x] += (T) v;

      return(*this);
    }

  template<class T> array<T>&		array<T>::operator*=(double v)
    {
      int		y, x;
      T            **_t;
   
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _t[y][x] = (T) (_t[y][x] * v);

      return(*this);
    }


  template<class T> array<T>&		array<T>::operator>(array<T>& A)
    {
      int		y, x;
      array<T>	*B;
      T            **_t, **_B;

      CheckArraysConsistency(*this, A);
   
      B	= new array<T>(A);
      _B   = (*B).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _B[y][x] = (_t[y][x] > _B[y][x]) ? 1 : 0;

      DeleteIfTemporary(*this);

      return((*B).temporary());
    }


  template<class T> array<T>&		array<T>::operator<(array<T>& A)
    {
      int		y, x;
      array<T>	*B;
      T            **_t, **_B;

      CheckArraysConsistency(*this, A);
   
      B	= new array<T>(A);
      _B   = (*B).storage();
      _t   = (*this).storage();
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _B[y][x] = (_t[y][x] < _B[y][x]) ? 1 : 0;

      DeleteIfTemporary(*this);

      return((*B).temporary());
    }



  //! Output operator. Array a is dumped in binary form to the given stream
  //! The format of the dump is the following:
  //!
  //! 		AnImAl header (32 bytes)
  //!		array elements (IEEE float) in a scanline order
  //!
  //! See file <io_heads> for the detailed structure of the header.
  //!

  

  
  template<class T> void array<T>::write(ostream&s, int fmt)
    {
      animal_header	header;
      int		x, y;
  
      header.MAGIC	= lIMAGE;
      header.rows	= nr_of_rows;
      header.cols	= nr_of_cols;
      header.focusx	= first_col_idx;
      header.focusy	= first_row_idx;
      header.outside= 0;
      header.flag	= 0;

      switch(fmt)
	{
	case anCHAR:
	  {
	    char value;
	  
	    header.type	= anCHAR;
	    fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);


	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (char)((*this)[y][x]);
		  s.write((char *) &value, sizeof(char));
		}
	  }
	  break;
	case anUCHAR:
	  {
	    unsigned char value;
	  
	    header.type	= anUCHAR;
	    fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);


	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (unsigned char)((*this)[y][x]);
		  s.write((char *) &value, sizeof(unsigned char));
		}
	  }
	  break;
	case anSHORT:
	  {
	    int value;
	  
	    header.type	= anSHORT;
	    fwritex((char *) &header, sizeof(short), sizeof(animal_header)/sizeof(int), s);


	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (int)((*this)[y][x]);
		  fwritex((char *) &value, sizeof(short), 1, s);
		}
	  }
	  break;
	case anINT:
	  {
	    int value;
	  
	    header.type	= anINT;
	    fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);


	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (int)((*this)[y][x]);
		  fwritex((char *) &value, sizeof(int), 1, s);
		}
	  }
	  break;
	case anFLOAT:
	  {
	    float value;
	  
	    header.type	= anFLOAT;
	    fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);

	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (float)((*this)[y][x]);
		  fwritex((char *) &value, sizeof(float), 1, s);
		}
	  }
	  break;
	case anDOUBLE:
	  {
	    double value;
	  
	    header.type	= anDOUBLE;
	    fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);

	    for(y = first_row_idx; y <= last_row_idx; y++)
	      for(x = first_col_idx; x <= last_col_idx; x++)
		{
		  value	= (double)((*this)[y][x]);
		  fwritex((char *) &value, sizeof(double), 1, s);
		}
	  }
	  break;
	default:
	  {
	    cerr << "Illegal type for formatted array dump\n";
	    exit(-1);
	  }
	}

      DeleteIfTemporary(*this);
    }


  /*
  // Old write function: force the cast to float
  template<class T> void array<T>::write(ostream&s)
  {
  animal_header	header;
  int		x, y;
  float		value;
  
  header.MAGIC	= lIMAGE;
  header.rows	= nr_of_rows;
  header.cols	= nr_of_cols;
  header.type	= anFLOAT;
  header.focusx	= first_col_idx;
  header.focusy	= first_row_idx;
  header.outside= 0;
  header.flag	= 0;
  
  fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);

  for(y = first_row_idx; y <= last_row_idx; y++)
  for(x = first_col_idx; x <= last_col_idx; x++)
  {
  value	= (*this)[y][x];
  fwritex((char *) &value, sizeof(float), 1, s);
  }

  DeleteIfTemporary(*this);
  }
  */

  
  template<class T> void array<T>::write(ostream&s)
    {
      animal_header	header;
      int		x, y;
      T 		value;
  
      int type =  AnimalType<T>();
      
      if(type == 0) 
	{
	  fprintf(stderr,"Array::write animal type not supported\n");
	  exit(-1);
	}

      header.MAGIC	= lIMAGE;
      header.rows	= nr_of_rows;
      header.cols	= nr_of_cols;
      header.type	= type;
      header.focusx	= first_col_idx;
      header.focusy	= first_row_idx;
      header.outside    = 0;
      header.flag	= 0;
  
      fwritex((char *) &header, sizeof(int), sizeof(animal_header)/sizeof(int), s);

      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  {
	    value	= (*this)[y][x];
	    fwritex((char *) &value, sizeof(T), 1, s);
	  }

      DeleteIfTemporary(*this);  
    }

  template<class T> void array<T>::write(char *filename)
    {
      ofstream out(filename);
   
      (*this).write(out);
   
      out.close();
    }

  template<class T> void array<T>::write(char *filename, int fmt)
    {
      ofstream out(filename);
   
      (*this).write(out, fmt);
   
      out.close();
    }



  template<class T> void array<T>::writeAscii()
    {
      int		  x, y;
      char           buf[1024];
   
      {
	for(y = first_row_idx; y <= last_row_idx; y++)
	  {
	    for(x = first_col_idx; x <= last_col_idx; x++)
	      {
		sprintf(buf, "%+18g ", float((*this)[y][x]));
		cerr << buf;
	      }
	    cerr << "\n";
	  }
      }
   
      DeleteIfTemporary(*this);
    }

  template<class T> void array<T>::writeAscii(ostream&s)
    {
      writeAscii(s,AnimalType<T>());
    }

  template<class T> void array<T>::writeAscii(ostream&s, int fmt)
    {
      int		  x, y;
      char           buf[1024];
   
      switch(fmt)
	{
	case anDOUBLE:
	  {
	    for(y = first_row_idx; y <= last_row_idx; y++)
	      {
		for(x = first_col_idx; x <= last_col_idx; x++)
		  {
		    sprintf(buf, "%+18g ", double((*this)[y][x]));
		    s << buf;
		  }
		s << "\n";
	      }
	  }
	  break;
	case anFLOAT:
	  {
	    for(y = first_row_idx; y <= last_row_idx; y++)
	      {
		for(x = first_col_idx; x <= last_col_idx; x++)
		  {
		    sprintf(buf, "%+18g ", float((*this)[y][x]));
		    s << buf;
		  }
		s << "\n";
	      }
	  }
	  break;
	case anINT:
	  {
	    for(y = first_row_idx; y <= last_row_idx; y++)
	      {
		for(x = first_col_idx; x <= last_col_idx; x++)
		  {
		    sprintf(buf, "%+11d ", int((*this)[y][x]));
		    s << buf;
		  }
		s << "\n";
	      }
	  }
	  break;
	case anCHAR:
	  {
	    for(y = first_row_idx; y <= last_row_idx; y++)
	      {
		for(x = first_col_idx; x <= last_col_idx; x++)
		  {
		    sprintf(buf, "%+4d ", char((*this)[y][x]));
		    s << buf;
		  }
		s << "\n";
	      }
	  }
	  break;
	case anUCHAR:
	  {
	    for(y = first_row_idx; y <= last_row_idx; y++)
	      {
		for(x = first_col_idx; x <= last_col_idx; x++)
		  {
		    sprintf(buf, "%+4d ", (unsigned char)((*this)[y][x]));
		    s << buf;
		  }
		s << "\n";
	      }
	  }
	  break;
	}
   
      DeleteIfTemporary(*this);
    }


  template<class T> void array<T>::writeAscii(char *filename,
					      int  fmt,
					      char *nr_format)
    {
      FILE           *out = NULL;
      int		  x, y;
      char           *std_float_format = "%+18g ";
      char           *std_int_format   = "%+11d ";
      char           *std_char_format  = "%+4d ";
      char           *std_uchar_format = "%+4d ";
      char           *out_format       = NULL;
   

      if(nr_format == NULL)
	switch(fmt)
	  {
	  case anDOUBLE:
	    out_format = std_float_format;
	    break;
	  case anFLOAT:
	    out_format = std_float_format;
	    break;
	  case anINT:
	    out_format = std_int_format;
	    break;
	  case anCHAR:
	    out_format = std_char_format;
	    break;
	  case anUCHAR:
	    out_format = std_uchar_format;
	    break;
	  }
      else
	out_format = nr_format;
   
   
      if(filename == NULL)
	out = stdout;
      else
	out = fopen(filename, "wb");
   
      if(out != NULL)
	{
	  switch(fmt)
	    {
	    case anDOUBLE:
	      {
		for(y = first_row_idx; y <= last_row_idx; y++)
		  {
		    for(x = first_col_idx; x <= last_col_idx; x++)
		      fprintf(out, out_format, double((*this)[y][x]));
		     
		    fprintf(out, "\n");
		  }
	      }
	      break;
	    case anFLOAT:
	      {
		for(y = first_row_idx; y <= last_row_idx; y++)
		  {
		    for(x = first_col_idx; x <= last_col_idx; x++)
		      fprintf(out, out_format, float((*this)[y][x]));
		     
		    fprintf(out, "\n");
		  }
	      }
	      break;
	    case anINT:
	      {
		for(y = first_row_idx; y <= last_row_idx; y++)
		  {
		    for(x = first_col_idx; x <= last_col_idx; x++)
		      fprintf(out, out_format, int((*this)[y][x]));

		    fprintf(out, "\n");
		  }
	      }
	      break;
	    case anCHAR:
	      {
		for(y = first_row_idx; y <= last_row_idx; y++)
		  {
		    for(x = first_col_idx; x <= last_col_idx; x++)
		      fprintf(out, out_format, char((*this)[y][x]));

		    fprintf(out, "\n");
		  }
	      }
	      break;
	    case anUCHAR:
	      {
		for(y = first_row_idx; y <= last_row_idx; y++)
		  {
		    for(x = first_col_idx; x <= last_col_idx; x++)
		      fprintf(out, out_format, (unsigned char)((*this)[y][x]));

		    fprintf(out, "\n");
		  }
	      }
	      break;
	    }

	  if(out != stdout)
	    fclose(out);
	}
   
      DeleteIfTemporary(*this);
    }



  //! Input operator. Array a is read in binary form from the given stream
  //! The format of the array is the following:
  //!
  //! 		AnImAl header (32 bytes)
  //!	    [	array elements (IEEE float) in a scanline order
  //!			|
  //!		array elements (unsigned char) in a scanline order
  //!			|
  //!		array elements (char) in a scanline order
  //!			|
  //!		array elements (short) in a scanline order
  //!			|
  //!		array elements (int) in a scanline order	]
  //!
  //! See file <io_heads> for the detailed structure of the header.
  //!

  template<class T> void array<T>::read(char *filename)
    {
      ifstream in(filename);
   
      if(in.is_open())
	{
	  (*this).read(in);
	}
      else
	{
	  fprintf(stderr,"Error reading file %s\n",filename);
	  exit(-1);
	}
   
      in.close();
    }

  template<class T> void array<T>::read(istream& s)
    {
      animal_header	header;
      int			x, y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			rows, cols;
      AnImAl_types		type;
      unsigned int		n;
      T                    **_A;
      float                old_scaling = scaling;
   
      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   

      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;
      n		= rows*cols;
   
      *this	= (*(new array<T>(low_y, high_y, low_x, high_x))).temporary();
      _A           = (*this).storage();
      (*this).scale(old_scaling);
   
      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;

	    buffer	= new u_char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(unsigned char));

		ptr = buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
		
	      }

	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;

	    buffer	= new char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(char));
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	    break;
	  }
	case anSHORT:
	  {
	    short *buffer, *ptr;

	    buffer	= new short[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(short), cols, s);

		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;

	    buffer	= new int[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(int), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(float), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(double), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}
    }   

  template<class T> void array<T>::read(istream& s, int i0, int i1)
    {
      animal_header	header;
      int			x, y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			rows, cols;
      int                  row_size = 0;
      AnImAl_types		type;
      T                    **_A;
      float                old_scaling = scaling;
   
      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   

      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;
   
      if((i0 < low_y) || (high_y < i1) || (i0 > i1))
	{
	  ArrayError("Illegal subsetting within array!");
	  exit(-1);
	}
      else
	{
	  switch(type)
	    {
	    case anCHAR:
	      row_size = cols * sizeof(char);
	      break;
	    case anUCHAR:
	      row_size = cols * sizeof(unsigned char);
	      break;
	    case anSHORT:
	      row_size = cols * sizeof(short);
	      break;
	    case anINT:
	      row_size = cols * sizeof(int);
	      break;
	    case anFLOAT:
	      row_size = cols * sizeof(float);
	      break;
	    case anDOUBLE:
	      row_size = cols * sizeof(double);
	      break;
	    }

	  s.seekg(sizeof(header) + row_size * (i0 - low_y));
	  low_y  = i0;
	  high_y = i1;
	}
   
      *this	= (*(new array<T>(low_y, high_y, low_x, high_x))).temporary();
      _A           = (*this).storage();
      (*this).scale(old_scaling);
   
      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;

	    buffer	= new u_char[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(unsigned char));
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;

	    buffer	= new char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(char));
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	    break;
	  }
	case anSHORT:
	  {
	    short *buffer, *ptr;

	    buffer	= new short[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(short), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;

	    buffer	= new int[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(int), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(float), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(double), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}
    }   

  //! Vector I must contain an ORDERED (!!!!!!) set of rows to be read. 
  //!

  template<class T> void array<T>::read(istream& s, vector<int> &I)
    {
      animal_header	header;
      int                  n = I.length();
      int			x, y;
      int                  row_size = 0;
      int                  off_y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			rows, cols;
      AnImAl_types		type;
      T                    **_A;
      float                old_scaling = scaling;
   
      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   
   
      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;

      //! Check the row numbers
      //!
      {
	int i;
	int i0      = 0;
	int i1      = n-1;
	int min_row = I[i0];
	int max_row = I[i0];
      
	for(i = i0+1; i <= i1; i++)
	  {
	    min_row = MIN(min_row, I[i]);
	    max_row = MAX(max_row, I[i]);
	  }
      
	if((min_row < low_y) || (high_y < max_row))
	  {
	    ArrayError("Illegal subsetting within array!");
	    exit(-1);
	  }
      }
   
      //! Determine the size of a single row
      //!
      switch(type)
	{
	case anCHAR:
	  row_size = cols * sizeof(char);
	  break;
	case anUCHAR:
	  row_size = cols * sizeof(unsigned char);
	  break;
	case anSHORT:
	  row_size = cols * sizeof(short);
	  break;
	case anINT:
	  row_size = cols * sizeof(int);
	  break;
	case anFLOAT:
	  row_size = cols * sizeof(float);
	  break;
	case anDOUBLE:
	  row_size = cols * sizeof(double);
	  break;
	}

      off_y  = low_y;
      low_y  = 0;
      high_y = n-1;
   
      *this	= (*(new array<T>(low_y, high_y, low_x, high_x))).temporary();
      _A           = (*this).storage();
      (*this).scale(old_scaling);
   
      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;
	    
	    buffer	= new u_char[cols];
	    
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		s.read((char *) buffer, cols*sizeof(unsigned char));
		 
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	    
	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;
	     
	    buffer	= new char[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		s.read((char *) buffer, cols*sizeof(char));
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	     
	    delete[] buffer;
	  }
	  break;
	case anSHORT:
	  {
	    short *buffer, *ptr;
	     
	    buffer	= new short[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(short), cols, s);
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	     
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;
	     
	    buffer	= new int[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(int), cols, s);
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(float), cols, s);
		
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr * scaling);
			ptr++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[y][x] = T(*ptr);
			ptr++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(double), cols, s);
		
		ptr	  	= buffer;
		for(x = low_x; x <= high_x; x++)
		  {
		    (_A)[y][x] = T(*ptr);
		    ptr++;
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}

    }

  template<class T> void array<T>::readAsSingleRow(istream& s)
    {
      animal_header	header;
      int			x, y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			idx, rows, cols;
      AnImAl_types		type;
      unsigned int		n;
      T                    **_A;
      float                old_scaling = scaling;
   
      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   

      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;
      n		= rows*cols;
   
      *this	= (*(new array<T>(1,n))).temporary();
      (*this).scale(old_scaling);
      _A           = (*this).storage();
      idx          = 0;

      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;

	    buffer	= new u_char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(unsigned char));

		ptr = buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }

	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;

	    buffer	= new char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(char));
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	    break;
	  }
	case anSHORT:
	  {
	    short *buffer, *ptr;

	    buffer	= new short[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(short), cols, s);

		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;

	    buffer	= new int[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(int), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(float), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(double), cols, s);
		
		ptr = buffer;
		
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}
    }   

  template<class T> void array<T>::readAsSingleRow(istream& s, int i0, int i1)
    {
      animal_header	header;
      int			x, y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			idx, rows, cols;
      int                  row_size = 0;
      AnImAl_types		type;
      int                  n;
      T                    **_A;

      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   

      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;
   
      if((i0 < low_y) || (high_y < i1) || (i0 > i1))
	{
	  ArrayError("Illegal subsetting within array!");
	  exit(-1);
	}
      else
	{
	  switch(type)
	    {
	    case anCHAR:
	      row_size = cols * sizeof(char);
	      break;
	    case anUCHAR:
	      row_size = cols * sizeof(unsigned char);
	      break;
	    case anSHORT:
	      row_size = cols * sizeof(short);
	      break;
	    case anINT:
	      row_size = cols * sizeof(int);
	      break;
	    case anFLOAT:
	      row_size = cols * sizeof(float);
	      break;
	    case anDOUBLE:
	      row_size = cols * sizeof(double);
	      break;
	    }

	  s.seekg(sizeof(header) + row_size * (i0 - low_y));
	  low_y  = i0;
	  high_y = i1;
	  n      = (high_y - low_y + 1) * cols;
	}

   
      *this	= (*(new array<T>(1,n))).temporary();
      _A           = (*this).storage();
      idx          = 0;
   
      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;

	    buffer	= new u_char[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(unsigned char));
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;

	    buffer	= new char[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		s.read((char *) buffer, cols*sizeof(char));
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	    break;
	  }
	case anSHORT:
	  {
	    short *buffer, *ptr;

	    buffer	= new short[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(short), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;

	    buffer	= new int[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(int), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(float), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];
	   
	    for(y = low_y; y <= high_y; y++)
	      {
		freadx(buffer, sizeof(double), cols, s);
		ptr	  	= buffer;

		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}
    }   

  //! Vector I must contain an ORDERED (!!!!!!) set of rows to be read. 
  //!

  template<class T> void array<T>::readAsSingleRow(istream& s, vector<int> &I)
    {
      animal_header	header;
      int                  n;
      int			x, y;
      int                  row_size = 0;
      int                  off_y;
      int			low_x, high_x;
      int			low_y, high_y;
      int			idx, rows, cols;
      AnImAl_types		type;
      T                    **_A;

      freadx((int *) &header,
	     sizeof(int),
	     sizeof(animal_header)/sizeof(int), s);   
   
      if((header.MAGIC != lIMAGE) && (header.MAGIC != lARRAY))
	{
	  cerr << "Bad magic number [" << header.MAGIC << "] for array!\n";
	  exit(-1);
	}

      type 	= AnImAl_types(header.type);
      low_x	= header.focusx;
      low_y	= header.focusy;
      rows		= header.rows;
      cols		= header.cols;
      high_x	= low_x + cols - 1;
      high_y	= low_y + rows - 1;

      //! Check the row numbers
      //!
      {
	int i;
	int i0      = 0;
	int i1      = I.length() - 1;
	int min_row = I[i0];
	int max_row = I[i0];
      
	for(i = i0+1; i <= i1; i++)
	  {
	    min_row = MIN(min_row, I[i]);
	    max_row = MAX(max_row, I[i]);
	  }
      
	if((min_row < low_y) || (high_y < max_row))
	  {
	    ArrayError("Illegal subsetting within array!");
	    exit(-1);
	  }
      }
   
      //! Determine the size of a single row
      //!
      switch(type)
	{
	case anCHAR:
	  row_size = cols * sizeof(char);
	  break;
	case anUCHAR:
	  row_size = cols * sizeof(unsigned char);
	  break;
	case anSHORT:
	  row_size = cols * sizeof(short);
	  break;
	case anINT:
	  row_size = cols * sizeof(int);
	  break;
	case anFLOAT:
	  row_size = cols * sizeof(float);
	  break;
	case anDOUBLE:
	  row_size = cols * sizeof(double);
	  break;
	}

      off_y  = low_y;
      low_y  = 0;
      high_y = I.length() - 1;
   
      n            = I.length() * cols; 
      *this	= (*(new array<T>(1,n))).temporary();
      _A           = (*this).storage();
      idx          = 0;
   
      switch(type)
	{
	case anUCHAR:
	  {
	    unsigned char *buffer, *ptr;
	    
	    buffer	= new u_char[cols];
	    
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		s.read((char *) buffer, cols*sizeof(unsigned char));
		 
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	    
	    delete[] buffer;
	  }
	  break;
	case anCHAR:
	  {
	    char *buffer, *ptr;
	     
	    buffer	= new char[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		s.read((char *) buffer, cols*sizeof(char));
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	     
	    delete[] buffer;
	  }
	  break;
	case anSHORT:
	  {
	    short *buffer, *ptr;
	     
	    buffer	= new short[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(short), cols, s);
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	     
	    delete[] buffer;
	  }
	  break;
	case anINT:
	  {
	    int *buffer, *ptr;
	     
	    buffer	= new int[cols];
	     
	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(int), cols, s);
		  
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anFLOAT:
	  {
	    float *buffer, *ptr;

	    buffer	= new float[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(float), cols, s);
		
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	case anDOUBLE:
	  {
	    double *buffer, *ptr;

	    buffer	= new double[cols];

	    for(y = low_y; y <= high_y; y++)
	      {
		s.seekg(sizeof(header) + row_size * (I[y]-off_y));
		freadx((char *) buffer, sizeof(double), cols, s);
		
		ptr	  	= buffer;
		if(scaling != 1.0)
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr * scaling);
			ptr++;
			idx++;
		      }
		  }
		else
		  {
		    for(x = low_x; x <= high_x; x++)
		      {
			(_A)[0][idx] = T(*ptr);
			ptr++;
			idx++;
		      }
		  }
	      }
	   
	    delete[] buffer;
	  }
	  break;
	default:
	  cerr << "Unsupported AnImAl type!\n";
	  exit(-1);
	}

    }

  //! Read an array stored as regular ascii format
  //!

  template<class T> void array<T>::readAscii(char *name)
    {
      FILE		        *input_file;
      unsigned int	        r    = 0;
      unsigned int	        c    = 0;
      unsigned int	        cs   = 0;
      char		        *idx;
      char		        num_buffer[128],  *num;
      static	char 	line_buffer[SIZE];
   
      input_file = fopen(name, "r");
      if(input_file == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
   
      while(fgets(line_buffer, SIZE-1, input_file) != NULL)
	{
	  if(line_buffer[0] != '#')
	    {
	      cs		= 0;
	      idx 	= line_buffer;
	     
	      while(*idx != '\n')
		{
		  while((*idx == ' ') || (*idx == '\t'))
		    idx++;
		  
		  if(*idx != '\n')
		    {
		      num = num_buffer;
		      while((*idx != ' ') && (*idx != '\n') && (*idx != '\t'))
			*num++ = *(idx++);
		       
		      *num = '\0';
		      cs++;
		    }
		}
	     
	      if(cs)
		{
		  r++;
		  if(!c)
		    c = cs;
		  else if(cs != c)
		    {
		      fprintf(stderr, "Parse error in %s!\n", name);
		      exit(-1);
		    }
		}
	    }
	}
   
      if((r > 0) && (c > 0))
	{
	  fseek(input_file, 0, 0);
	  *this	= (*(new array<T>(r, c))).temporary();
	
	  r	= 0;
	  while(fgets(line_buffer, SIZE-1, input_file) != NULL)
	    {
	      if(line_buffer[0] != '#')
		{
		  cs		= 0;
		  idx 	= line_buffer;
		  
		  while(*idx != '\n')
		    {
		      while((*idx == ' ') || (*idx == '\t'))
			idx++;
		       
		      if(*idx != '\n')
			{
			  num = num_buffer;
			  while((*idx!=' ')&&(*idx != '\n')&&(*idx!='\t'))
			    *num++ = *(idx++);
			    
			  *num = '\0';
			  (*this)[r][cs] = (T)(atof(num_buffer));
			  cs++;
			}
		    }
		  
		  if(cs)
		    {
		      r++;
		      if(!c)
			c = cs;
		      else if(cs != c)
			{
			  fprintf(stderr, "Parse error in %s!\n", name);
			  exit(-1);
			}
		    }
		}
	    }
	}
      else
	{
	  fprintf(stderr, "No valid data in %s!\n", name);
	  exit(-1);
	}
   
      fclose(input_file);
    }

  template<class T> void array<T>::readAscii(char *name, int i0, int i1)
    {
      FILE		        *input_file;
      int	                r    = 0;
      unsigned int	        c    = 0;
      unsigned int	        cs   = 0;
      char		        *idx;
      char		        num_buffer[128],  *num;
      static	char 	line_buffer[SIZE];
   
      input_file = fopen(name, "r");
      if(input_file == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
   
      //! Read the file up to the first line with
      //! data to get the number of columns
      //!
      while((fgets(line_buffer, SIZE-1, input_file) != NULL) && (r == 0))
	{
	  if(line_buffer[0] != '#')
	    {
	      cs		= 0;
	      idx 	= line_buffer;
	     
	      while(*idx != '\n')
		{
		  while((*idx == ' ') || (*idx == '\t'))
		    idx++;
		  
		  if(*idx != '\n')
		    {
		      num = num_buffer;
		      while((*idx != ' ') && (*idx != '\n') && (*idx != '\t'))
			*num++ = *(idx++);
		       
		      *num = '\0';
		      cs++;
		    }
		}
	     
	      if(cs)
		{
		  r++;
		  if(!c)
		    c = cs;
		  else if(cs != c)
		    {
		      fprintf(stderr, "Parse error in %s!\n", name);
		      exit(-1);
		    }
		}
	    }
	}
   
      //! Go to start of file
      //!
      fseek(input_file, 0, 0);

   
      *this	= (*(new array<T>(i0, i1, 0, c-1))).temporary();
   
      r	= -1;
      while(fgets(line_buffer, SIZE-1, input_file) != NULL)
	{
	  if(line_buffer[0] != '#')
	    {
	      r++;
	     
	      if(r >= i0)
		{
		  
		  cs		= 0;
		  idx 	= line_buffer;
	     
		  while(*idx != '\n')
		    {
		      while((*idx == ' ') || (*idx == '\t'))
			idx++;
		  
		      if(*idx != '\n')
			{
			  num = num_buffer;
			  while((*idx!=' ')&&(*idx != '\n')&&(*idx!='\t'))
			    *num++ = *(idx++);
			    
			  *num = '\0';
			  (*this)[r][cs] = (T)(atof(num_buffer));
			  cs++;
			}
		    }
		  
		  if(cs != c)
		    {
		      fprintf(stderr, "Parse error <empty line> in %s!\n", name);
		      exit(-1);
		    }
		}
	    }

	  if(r == i1)
	    break;
	}
   
      fclose(input_file);
    }

  template<class T> void array<T>::readAscii(char *name, vector<int> &I)
    {
      FILE		        *input_file;
      int                  n    = I.length();
      int	                r    = 0, i0, i1;
      unsigned int	        c    = 0;
      unsigned int	        cs   = 0;
      char		        *idx;
      char		        num_buffer[128],  *num;
      int                  m;
      static	char 	line_buffer[SIZE];
   
      input_file = fopen(name, "r");
      if(input_file == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
   
      //! Read the file up to the first line with
      //! data to get the number of columns
      //!
      while((fgets(line_buffer, SIZE-1, input_file) != NULL) && (r == 0))
	{
	  if(line_buffer[0] != '#')
	    {
	      cs		= 0;
	      idx 	= line_buffer;
	     
	      while(*idx != '\n')
		{
		  while((*idx == ' ') || (*idx == '\t'))
		    idx++;
		  
		  if(*idx != '\n')
		    {
		      num = num_buffer;
		      while((*idx != ' ') && (*idx != '\n') && (*idx != '\t'))
			*num++ = *(idx++);
		       
		      *num = '\0';
		      cs++;
		    }
		}
	     
	      if(cs)
		{
		  r++;
		  if(!c)
		    c = cs;
		  else if(cs != c)
		    {
		      fprintf(stderr, "Parse error in %s!\n", name);
		      exit(-1);
		    }
		}
	    }
	}
   
      //! Allocate array
      //!
      *this	= (*(new array<T>(0, n-1, 0, c-1))).temporary();

      //! Go to start of file
      //!
      fseek(input_file, 0, 0);

      r	= -1;
      m    = 0;
      i0   = I[m];
      i1   = I[n-1];
   
      while(fgets(line_buffer, SIZE-1, input_file) != NULL)
	{
	  if(line_buffer[0] != '#')
	    {
	      r++;
	     
	      if(r == i0)
		{
		  cs	= 0;
		  idx 	= line_buffer;
	     
		  while(*idx != '\n')
		    {
		      while((*idx == ' ') || (*idx == '\t'))
			idx++;
		  
		      if(*idx != '\n')
			{
			  num = num_buffer;
			  while((*idx!=' ')&&(*idx != '\n')&&(*idx!='\t'))
			    *num++ = *(idx++);
			    
			  *num = '\0';
			  (*this)[m][cs] = (T)(atof(num_buffer));
			  cs++;
			}
		    }
		  
		  if(cs != c)
		    {
		      fprintf(stderr, "Parse error <empty line> in %s!\n", name);
		      exit(-1);
		    }

		  //! Increment index within indexing array
		  //!
		  m++;
		  
		  //! and set new target row
		  //!
		  if(m < n)
		    i0    = I[m];
		}
	    }

	  if(r == i1)
	    break;
	}
   
      fclose(input_file);
    }

  template<class T> void array<T>::get(char *name)
    {
      FILE *in = NULL;
      int  magic = 0;
   
      if((in = fopen(name, "rb")) == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
      else
	{
	  freadx((int *) &magic, sizeof(int), 1, in);
	  fclose(in);

	  if((magic == lIMAGE) || (magic == lARRAY))
	    {
	      ifstream data(name);
	     
	      (*this).read(data);

	      data.close();
	    }
	  else
	    {
	      (*this).readAscii(name);
	    }
	}
    }
    
  template<class T> void array<T>::get(char *name, int r0 , int r1)
    {
      FILE *in = NULL;
      int  magic = 0;
   
      if((in = fopen(name, "rb")) == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
      else
	{
	  freadx((int *) &magic, sizeof(int), 1, in);
	  fclose(in);

	  if((magic == lIMAGE) || (magic == lARRAY))
	    {
	      ifstream data(name);
	     
	      (*this).read(data, r0, r1);

	      data.close();
	    }
	  else
	    {
	      (*this).readAscii(name, r0, r1);
	    }
	}
    }

  template<class T> void array<T>::get(char *name, vector<int> &I)
    {
      FILE *in = NULL;
      int  magic = 0;
   
      if((in = fopen(name, "rb")) == NULL)
	{
	  fprintf(stderr, "Cannot open %s for reading!\n", name);
	  exit(-1);
	}
      else
	{
	  freadx((int *) &magic, sizeof(int), 1, in);
	  fclose(in);

	  if((magic == lIMAGE) || (magic == lARRAY))
	    {
	      ifstream data(name);
	     
	      (*this).read(data, I);

	      data.close();
	    }
	  else
	    {
	      (*this).readAscii(name, I);
	    }
	}
    }

  //! Return as a new array the <A>^n.
  //! The results is tagged for deletion
  //!

  template<class T> array<T>&	array<T>::operator^(int n)
    {
      
   
      if(n < 0)
	{
	  return(Inverse(*this)^(-n));
	}
      else
	{
	  if(nr_of_rows != nr_of_cols)
	    {
	      cerr << "array::operator^: array must be square!\n";
	      exit(-1);
	    }
   
	  array<T> * X = NULL;

	  if(n == 0)
	    {
	       
	      X = new array<T>(first_row_idx, last_row_idx, first_col_idx, last_col_idx);
	      T ** _X = X->storage();

	      for(int y = first_row_idx; y <= last_row_idx; y++)
		_X[y][y] = 1;

	    }
	  else 
	    {
	      X = new array<T>(*this);
	    	     
	      for(int i = 2; i <= n; i++)
		(*X) = (*X) * (*this);
	   
	    }
   
	  return((*X).temporary());
	}
    }

  template<class T> array<T>& array<T>::operator||(array<T>& B)
    {
      CheckArraysConsistency(B, *this);
   
      array<T>* OUT = new array<T>(first_row_idx, last_row_idx, first_col_idx, last_col_idx);
   
      T** _OUT = OUT->storage();
      T** _A = this->storage();
      T** _B = B.storage();
      T *o, *a, *b;
   
      for(int y = first_row_idx; y <= last_row_idx; y++)
	{
	  a = _A[y];
	  b = _B[y];
	  o = _OUT[y];

	  for(int x = first_col_idx; x <= last_col_idx; x++)
	    o[x] = a[x] || b[x];
	}

      DeleteIfTemporary(*this);
      DeleteIfTemporary(B);
   
      return OUT->temporary();
    }


  template<class T> array<T>& array<T>::operator&&(array<T>& B)
    {
      CheckArraysConsistency(B, *this);
   
      array<T>* OUT = new array<T>(first_row_idx, last_row_idx, first_col_idx, last_col_idx);
   
      T** _OUT = OUT->storage();
      T** _A = this->storage();
      T** _B = B.storage();
      T *o, *a, *b;
   
      for(int y = first_row_idx; y <= last_row_idx; y++)
	{
	  a = _A[y];
	  b = _B[y];
	  o = _OUT[y];

	  for(int x = first_col_idx; x <= last_col_idx; x++)
	    o[x] = a[x] && b[x];
	}

      DeleteIfTemporary(*this);
      DeleteIfTemporary(B);
   
      return OUT->temporary();
    }


  template<class T> array<T>& array<T>::operator!()
    {
      array<T>* OUT = new array<T>(first_row_idx, last_row_idx, first_col_idx, last_col_idx);
   
      T** _OUT = OUT->storage();
      T** _A = this->storage();
      T *o, *a;
   
      for(int y = first_row_idx; y <= last_row_idx; y++)
	{
	  a = _A[y];
	  o = _OUT[y];

	  for(int x = first_col_idx; x <= last_col_idx; x++)
	    o[x] = !a[x];
	}

      DeleteIfTemporary(*this);
   
      return OUT->temporary();
    }

  //============================================================================
  //FUNCTIONS




  //! Apply the abs function to each element
  //! The result is tagged for deletion
  //!

  template<class T> array<T>& abs(array<T>& A)
    {
      int		y, x;
      int          first_row_idx, last_row_idx;
      int          first_col_idx, last_col_idx;
      array<T>	*X;
      T            **_X;

      first_row_idx = A.first_row_index();
      last_row_idx  = A.last_row_index();
      first_col_idx = A.first_col_index();
      last_col_idx  = A.last_col_index();
   
      X	= new array<T>(A);
      _X   = (*X).storage();

      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = (_X[y][x] > 0.0) ? _X[y][x] : -_X[y][x];
   
      return((*X).temporary());
    }



  //! Apply the log function to each element
  //! The result is tagged for deletion
  //!

  template<class T> array<T>& log(array<T>& A)
    {
      int		y, x;
      int          first_row_idx, last_row_idx;
      int          first_col_idx, last_col_idx;
      array<T>	*X;
      T            **_X;

      first_row_idx = A.first_row_index();
      last_row_idx  = A.last_row_index();
      first_col_idx = A.first_col_index();
      last_col_idx  = A.last_col_index();
   
      X	= new array<T>(A);
      _X   = (*X).storage();

      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = log(_X[y][x]);
   
      return((*X).temporary());
    }



  //! Apply the sqrt function to each element
  //! The result is tagged for deletion
  //!

  template<class T> array<T>& sqrt(array<T>& A)
    {
      int		y, x;
      int          first_row_idx, last_row_idx;
      int          first_col_idx, last_col_idx;
      array<T>	*X;
      T            **_X;

      first_row_idx = A.first_row_index();
      last_row_idx  = A.last_row_index();
      first_col_idx = A.first_col_index();
      last_col_idx  = A.last_col_index();
   
      X	= new array<T>(A);
      _X   = (*X).storage();

      for(y = first_row_idx; y <= last_row_idx; y++)
	for(x = first_col_idx; x <= last_col_idx; x++)
	  _X[y][x] = T(sqrt(_X[y][x]));
   
      return((*X).temporary());
    }



  //! Return as a new array the transpose of <A>
  //! The results is tagged for deletion
  //!

  template<class T> array<T>&	Tr(array<T> &A)
    {
      array<T>	*X;
      int		i, j;
   
      X 		= new array<T>(A.first_col_index(), A.last_col_index(),
				       A.first_row_index(), A.last_row_index());
   
      for(j = A.first_row_index(); j <= A.last_row_index();  j++)
	{
	  for(i = A.first_col_index(); i <= A.last_col_index(); i++)
	    (*X)[i][j] = A[j][i];
	}
   

      DeleteIfTemporary(A);
   
      return((*X).temporary());
    }

  //! Return I smoothed according to the specified sigma
  //!
  template<class T> array<T>&	GaussianSmooth(array<T>& I, double sigma)
    {
      array<T>	*G;
      vector<T>    V;
      double	*K, n, v;
      T            **_I, **_G;
      int		w;
      int		i, y, x;
      int          r0, r1;
      int          c0, c1;

      if(sigma == 0.0)
		{
		  G  = new array<T>(I);
		}
      else
		{
		  T      *_i, *_V;
		  int     h = MAX(I.cols(), I.rows());
		  int     I0[h], *_i0;
		  int     DI[h], *_di;
		  int     j = 0, i0, di;
		  double  NN[h], *_nn;
		  double *KK[h],**_kk;
		  double *_K;

		  w    = int(floor(3*sigma+0.5));
		  K    = new double[2*w+1];

		  I.stat(c0, c1, r0, r1);
   
		  for(i = -w; i <= w; i++)
			K[i+w] 	= exp(-(i*i)/(2*sigma*sigma));
   
		  G            = new array<T>(r0, r1, c0, c1);
		  V            = (* new vector<T>(r0, r1)).temporary();
		  _G           = (*G).storage();
		  _V           = V.storage();
		  _I           = I.storage();


		  // Precompute normalizaton factors and convolution 
		  // window limits
		  // 
		  for(j = 0, x = c0; x <= c1; x++, j++)
			{
			  I0[j] = i0 = MAX(c0,x-w);
			  DI[j] = i  = MIN(c1, x+w) - i0 + 1;
	    
			  _K    = &(K[i0-x+w]);
			  KK[j] = _K;

			  NN[j]    = 0.0;
			  while(i--)
				NN[j] += *_K++;
			}	

		  for(y = r0; y <= r1; y++)
			{
			  _nn = &NN[0]; 
			  _i0 = &I0[0];
			  _di = &DI[0];
			  _kk = &KK[0];

			  for(x = c0; x <= c1; x++)
				{
				  v  = 0.0;
				  i0 = *_i0++;
				  di = *_di++;
				  n  = *_nn++;

				  _K = *_kk++;
				  _i = &(_I[y][i0]);

				  while(di--)
					v += *_i++ * *_K++;

				  _G[y][x] = (T)(v / n);
				}	
			}

		  for(j = 0, y = r0; y <= r1; y++, j++)
			{
			  I0[j] = i0 = MAX(r0,y-w);
			  DI[j] = i  = MIN(r1,y+w) - i0 + 1;
	    
			  _K    = &(K[i0-y+w]);
			  KK[j] = _K;

			  NN[j] = 0.0;
			  while(i--)
				NN[j] += *_K++;
			}	

		  // Smooth over rows 
		  //
		  for(x = c0; x <= c1; x++)
			{
			  _nn = &NN[0]; 
			  _i0 = &I0[0];
			  _di = &DI[0];
			  _kk = &KK[0];

			  // Copy row into buffer
			  //
			  for(y = r0; y <= r1; y++)
				_V[y] = _G[y][x];
	     
			  for(y = r0; y <= r1; y++)
				{
				  v  = 0.0;
				  i0 = *_i0++;
				  di = *_di++;
				  n  = *_nn++;

				  _K = *_kk++;
	       
				  _i = &(_V[i0]);
				  while(di--)
					v += *_i++ * *_K++;

				  _G[y][x] = (T)(v / n);
				}	
			}

		  DeleteIfTemporary(I);
		  delete[] K;
		}

      // fprintf(stderr, "Hello ");
   
      return((*G).temporary());
    }
    
  template<class T> array<T>&	GaussianSmooth2(array<T>& I, double sigma)
    {
      array<T>	*G, *GG;
      double	*K, n, v;
      T            **_I, **_GG;
      int		w;
      int		i, y, x;
      int          r0, r1;
      int          c0, c1;

      if(sigma == 0.0)
		{
		  G  = new array<T>(I);
		}
      else
		{
		  w    = int(floor(3*sigma+0.5));
		  K    = new double[2*w+1];

		  r0   = I.first_row_index();
		  r1   = I.last_row_index();
   
		  c0   = I.first_col_index();
		  c1   = I.last_col_index();
   
		  for(i = -w; i <= w; i++)
			K[i+w] 	= exp(-(i*i)/(2*sigma*sigma));
   
		  G            = new array<T>(r0, r1, c0, c1);
		  GG           = new array<T>(r0, r1, c0, c1);
		  _I           = I.storage();
		  _GG          = (*GG).storage();
	
		  for(y = r0; y <= r1; y++)
			for(x = c0; x <= c1; x++)
			  {
				int      i0, i1;
				// double  *KK = K + w + i0 - x;
				v	= 0.0;
				n	= 0.0;

				i0 = MAX(c0,x-w);
				i1 = MIN(c1, x+w);
	       
				for(i = i0; i <= i1; i++)
				  {
					v += _I[y][i] * K[i-x+w];
					n += K[i-x+w];
				  }

				_GG[y][x] = (T)(v / n);
			  }	
 
		  for(x = c0; x <= c1; x++)
			for(y = r0; y <= r1; y++)
			  {
				int      i0, i1;

				v	= 0.0;
				n	= 0.0;

				i0 = MAX(r0, y-w);
				i1 = MIN(r1, y+w);
	       
				for(i = i0; i <= i1; i++)
				  {
					v += _GG[i][x] * K[i-y+w];
					n += K[i-y+w];
				  }

				(*G)[y][x] = (T)(v / n);
			  }	
   

		  DeleteIfTemporary(I);
		  delete[] K;
		  delete GG;
		}
   
      return((*G).temporary());
    }



  //! Reduce image by pixel subsampling
  //!

  template<class T> array<T>&	Unzoom(array<T>& AA, int sx, int sy)
    {
      array<T>	*R, A = AA;
      int		rr, cc, r, c;
      int		x, y, xx, yy;
      int          x0 = A.first_col_index();
      int          y0 = A.first_row_index();
      T            **_R, **_A;
   
      A.translate_at(0,0);
   
      rr 	= A.rows();
      cc 	= A.cols();
      r    = rr/sy;
      c    = cc/sx;
	   
      R	= new array<T>(r,c);

      _A  = A.storage();
      _R  = (*R).storage();

      for(y = 0, yy = 0; (yy < r) && (y < rr); yy++, y += sy)
	for(x = 0, xx = 0; (xx < c) && (x < cc); xx++, x += sx)
	  _R[yy][xx] = _A[y][x];

      (*R).translate_at(y0 / sy, x0 / sx);

      return((*R).temporary());
    }


  //! Reduce image by pixel subsampling
  //!

  template<class T> array<T>&	Linearize(array<T>& A)
    {
      array<T>	*R;
      T            **_R, **_A;
      int          y, first_row_idx, last_row_idx;
      int          x, first_col_idx, last_col_idx;
      int          size, idx;
   

      first_row_idx = A.first_row_index();
      last_row_idx  = A.last_row_index();
      first_col_idx = A.first_col_index();
      last_col_idx  = A.last_col_index();
  
      size = (last_row_idx - first_row_idx + 1) * (last_col_idx - first_col_idx + 1);
   
      R    = new array<T>(1,size);
      _R   = (*R).storage();
      _A   = A.storage();
   
      idx  = 0;
   
      for(y = first_row_idx; y <= last_row_idx; y++)
	{
	  for(x = first_col_idx; x <= last_col_idx; x++)
	    {
	      (*R)[0][idx] = _A[y][x];
	      idx++;
	    }
	}

      DeleteIfTemporary(A);
   
      return((*R).temporary());
    }


  template<class T> array<T>&	Zoom(array<T>& AA, int sx, int sy)
    {
      array<T>	*R, A = AA;
      int		rr, cc;
      int		x, y, zx, zy, i, j;
      int          x0 = A.first_col_index();
      int          y0 = A.first_row_index();
      T            **_R, **_A;

      A.translate_at(0,0);
   
      rr 		= A.rows();
      cc 		= A.cols();
   
      R   = new array<T>(rr*sy, cc*sx);

      _A  = A.storage();
      _R  = (*R).storage();
   
      for(y = 0; y < rr; y++)
	{
	  zy = sy * y;
	  for(x = 0; x < cc; x++)
	    {
	      zx = sx * x;
	      for(j =  0; j < sy; j++)
		{
		  for(i = 0; i < sx; i++)
		    {
		      _R[zy+j][zx+i] = _A[y][x];
		    }
		}
	    }
	}

      (*R).translate_at(y0 * sy, x0 * sx);

      return((*R).temporary());
    }

  //! Returns an image whose pixel values are set to 1 whenever a pixel
  //! greater than zero is found in the orignal image within a distance
  //! (city-block type) of n
  //!

  template<class T> array<T> &Grow(array<T> &D, int n)
    {
      array<T>	*A;
      int		x0, x1, y0, y1;
      int		x, y, i, j;
      int		sy, ey, sx, ex;
      T            v;
      T            **_A, **_D;
   
   
      x0   = D.first_col_index();
      x1   = D.last_col_index();
      y0   = D.first_row_index();
      y1   = D.last_row_index();
   
      A    = new array<T>(y0, y1, x0, x1);

      _A  = (*A).storage();
      _D  = D.storage();

      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    {
	      /* Compute legal limits */
	      sy = MAX(y-n, y0);
	      ey = MIN(y+n, y1);
	      sx = MAX(x-n, x0);
	      ex = MIN(x+n, x1);
	     
	      /* Grow on n-neighborhood */
	      v = _D[y][x];
	      for(j = sy; (j <= ey); j++)
		for(i = sx; (i <= ex); i++)
		  v = MAX(v, _D[j][i]);

	      _A[y][x] = v;
	    }
	}

      DeleteIfTemporary(D);
   
      return((*A).temporary());
    }


  template<class T> array<T> &Shrink(array<T> &D, int n)
    {
      array<T>	*A;
      int		x0, x1, y0, y1;
      int		x, y, i, j;
      int		sy, ey, sx, ex;
      T            v;
      T            **_A, **_D;
   
   
      x0   = D.first_col_index();
      x1   = D.last_col_index();
      y0   = D.first_row_index();
      y1   = D.last_row_index();
   
      A    = new array<T>(y0, y1, x0, x1);

      _A  = (*A).storage();
      _D  = D.storage();

      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    {
	      /* Compute legal limits */
	      sy = MAX(y-n, y0);
	      ey = MIN(y+n, y1);
	      sx = MAX(x-n, x0);
	      ex = MIN(x+n, x1);
	     
	      /* Grow on n-neighborhood */
	      v = _D[y][x];
	      for(j = sy; (j <= ey); j++)
		for(i = sx; (i <= ex); i++)
		  v = MIN(v, _D[j][i]);

	      _A[y][x] = v;
	    }
	}

      DeleteIfTemporary(D);
   
      return((*A).temporary());
    }


  //structuring element: 2m+1-height, 2n+1-width rectangle
  template<class T> array<T> & Grow(array<T> &D, int m, int n)
    {
      array<T>	*A;
      int		x0, x1, y0, y1;
      int		x, y, i, j;
      int		sy, ey, sx, ex;
      T            v;
      T            **_A, **_D;
   
   
      x0   = D.first_col_index();
      x1   = D.last_col_index();
      y0   = D.first_row_index();
      y1   = D.last_row_index();
   
      A    = new array<T>(y0, y1, x0, x1);

      _A  = (*A).storage();
      _D  = D.storage();

      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    {
	      /* Compute legal limits */
	      sy = MAX(y-m, y0);
	      ey = MIN(y+m, y1);
	      sx = MAX(x-n, x0);
	      ex = MIN(x+n, x1);
	     
	      /* Grow on n-neighborhood */
	      v = _D[y][x];
	      for(j = sy; (j <= ey); j++)
		for(i = sx; (i <= ex); i++)
		  v = MAX(v, _D[j][i]);

	      _A[y][x] = v;
	    }
	}

      return((*A).temporary());
    }


  //structuring element: 2m+1-height, 2n+1-width rectangle
  template<class T> array<T> &Shrink(array<T> &D, int m, int n)
    {
      array<T>	*A;
      int		x0, x1, y0, y1;
      int		x, y, i, j;
      int		sy, ey, sx, ex;
      T            v;
      T            **_A, **_D;
   
   
      x0   = D.first_col_index();
      x1   = D.last_col_index();
      y0   = D.first_row_index();
      y1   = D.last_row_index();
   
      A    = new array<T>(y0, y1, x0, x1);

      _A  = (*A).storage();
      _D  = D.storage();

      for(y = y0; y <= y1; y++)
	{
	  for(x = x0; x <= x1; x++)
	    {
	      /* Compute legal limits */
	      sy = MAX(y-m, y0);
	      ey = MIN(y+m, y1);
	      sx = MAX(x-n, x0);
	      ex = MIN(x+n, x1);
	     
	      /* Grow on n-neighborhood */
	      v = _D[y][x];
	      for(j = sy; (j <= ey); j++)
		for(i = sx; (i <= ex); i++)
		  v = MIN(v, _D[j][i]);

	      _A[y][x] = v;
	    }
	}

      return((*A).temporary());
    }


  

 

  //! Check whether two arrays differ in base index  and/or dimensions.
  //! Should they differ, the program exits.  
  //!

  template<class T, class U> void CheckArraysConsistency(const array<T> &A, const array<U> &B)
    {
      if((A.first_row_index() != B.first_row_index())	||
	 (A.first_col_index() != B.first_col_index())	||
	 (A.rows()		   != B.rows())			||
	 (A.cols()		   != B.cols()))
	{
	  cerr << "Arrays differ in base indexes and/or dimensions!\n";
	  exit(-1);
	}
    }


  //! Delete an array  if it was tagged for deletion
  //!

  template<class T> inline void DeleteIfTemporary(array<T>& A)
    {
      if(A.dispose() == TRUE)
	delete &A;
    }


  //! Cast operator: copy first argument into an array with the same type
  //! of the second argument
  //!

  template<class T, class U> array<T>& Cast(array<U>& B, array<T>& A)
    {
      int          x, x0, x1;
      int          y, y0, y1;
      
      x0  = B.first_col_index();
      x1  = B.last_col_index();
      y0  = B.first_row_index();
      y1  = B.last_row_index();
   
      array<T> * X  = new array<T>(y0, y1, x0, x1);
      
      T ** _X = X->storage();
      U ** _B = B.storage();
      T * __X;
      U * __B;

      for(y = y0; y <= y1; y++)
	{
	  __X = _X[y] + x0;
	  __B = _B[y] + x0;
	  for(x = x0; x <= x1; x++)
	    {
	      *__X++ = T( *__B++);
	    }
	}
   
      DeleteIfTemporary(A);
      DeleteIfTemporary(B);
   
      return((*X).temporary());
    }
  
  //! Cast operator
  //! Usage:
  //! CastTo<int>(A); // return an array of int

  template<class T, class U>
    array<T> & CastTo(array<U> & A)
    {
      int          x;
      int          y;
    
      
      int x0  = A.first_col_index();
      int x1  = A.last_col_index();
      int y0  = A.first_row_index();
      int y1  = A.last_row_index();
   
      array<T> * X  = new array<T>(y0, y1, x0, x1);

      T ** _X = X->storage();
      U ** _A = A.storage();
      T * __X;
      U * __A;

      for(y = y0; y <= y1; y++)
	{
	  __X = _X[y] + x0;
	  __A = _A[y] + x0;
	  for(x = x0; x <= x1; x++)
	    {
	      *__X++ = T(*__A++);
	    }
	}
      DeleteIfTemporary(A);
     
      return((*X).temporary());  
    }
  

  template<class T> array<T> & Inverse(array<T> & A)
    {
      int		        *indx;
      int		        n = A.rows();
      int		        i, j;

      double*	        col;
      double*               b;
      double	        d;

      array<double>         AA(CastTo<double>(A));
      array<double>         B; //LUD decomposition	
     
      double**              _A;
      double**              _B;
    
      if (n != A.cols())
	{
	  fprintf(stderr,"Inverse:\nCannot invert rectangular matrix!\n");
	  exit(-1);
	}
  
      // Allocate vector to keep track of row permutation
      indx	= new int[n];

      // Allocate aux vectors
      col = new double[n];
      b   = new double[n];
      
      B  = AA;
      _A = AA.storage();
      _B = B.storage();
    

      // Compute LU decomposition 
      if( ludcmp(_B, indx, n, &d) == FALSE )
	{
	  return (new array<T>)->temporary();
	}

      array<T> *I = new array<T>(n,n);
      T ** _I = I->storage();

      for(j = 0; j < n; j++)
	{
      
	  // Fill the j-th column 
	  for(i = 0; i < n; i++)
	    {
	      b[i]   = 0.0;
	      col[i] = 0.0;
	    }
	  b[j]   = 1.0;
	  col[j] = 1.0;
      
	  lubksb(_B, indx, col, n);
  
	  //Solution refinement    
	  mprove(_A, _B, indx, b, col, n);
	  mprove(_A, _B, indx, b, col, n);
     
	  //Copy the solution
	  for(i = 0; i < n; i++)
	    _I[i][j] = (T) col[i];
	}
    
      delete[] indx;
      delete[] col;
      delete[] b;

      return I->temporary();
    }

  template<class T> double	Det(array<T>& A)
    {
      int		        *indx;
      int		        n;
      int		        j;
      double	        d;
      array<double>	B;		//! Copy array to be inverted 
      
   
      //! An array<double> is required
      //!
      B = CastTo<double>(A);
      double ** _B = B.storage();
      
      //! Allocate vector to keep track of row permutation
      n     = B.rows();
      indx  = new int[n];
      
      //! Compute LU decomposition 
      ludcmp(_B, indx, n, &d);
   
      //! Ccompute determinant
      for(j = 0; j < n; j++)
	d  *= B[j][j];
   
      delete[] indx;
   
      return(d);
    }

}

#endif
