#ifndef	_array_iter_h
#define _array_iter_h 1


#include <arrayTools/namespace_header.h>


#define SIZE 128000

namespace _ARRAYTOOLS_NAMESPACE
{

  template<class T> class array;
 

  template<class T> 
    class _array_iterator_base
    { 
    public:

    
      array<T> * _array;
      int row, col; 
      int r0, r1, c0, c1;
      T **       _storage;     

      _array_iterator_base(array<T> * A, int Row, int Col) : _array(A), row(Row), col(Col) 
	{
	  _storage = A->storage();
	  A->boundary(r0,r1,c0,c1);
	}
      _array_iterator_base(array<T> * A, 
			   int Row, int Col, 
			   int R0, int R1, int C0, int C1) : _array(A), row(Row), col(Col), r0(R0), r1(R1), c0(C0), c1(C1) 
	{
	  _storage = A->storage();
	}
      _array_iterator_base() {}


      //Warning: no checks on increment decrement

      void _incr()
	{
	  if (++col > c1)
	    {
	      col = c0;
	      ++row;
	    }
	}

      void _decr()
	{
	  if (--col < c0)
	    {
	      col = c1;
	      --row;
	    }
	}

      bool operator==(const _array_iterator_base & x) const
	{
	  return (_array   == x._array   &&
		  row      == x.row      &&
		  col      == x.col);
	}

      bool operator!=(const _array_iterator_base& x) const
	{
	  return !(this->operator==(x));
	}

      void get_point(int & r, int & c) 
	{
	  r = row;
	  c = col;
	}

      T & point() { return _storage[row][col]; };
    };

  



  template <class Tp, class Ref, class Ptr>
    class _array_iterator : 
    public _array_iterator_base<Tp>, 
    public std::iterator<std::random_access_iterator_tag,Tp,ptrdiff_t, Ptr,Ref>
    {
    public:

      typedef _array_iterator<Tp, Tp&, Tp*>             iterator;
      typedef _array_iterator<Tp, const Tp&,const Tp*>  const_iterator;
      typedef _array_iterator<Tp, Ref, Ptr>             self;

      typedef Tp value_type;
      typedef Ptr pointer;
      typedef Ref reference;


      _array_iterator(array<Tp> * A, int Row, int Col)
	: _array_iterator_base<Tp>(A, Row, Col) {} 
      _array_iterator(array<Tp> * A, int Row, int Col, int R0, int R1, int C0, int C1)
	: _array_iterator_base<Tp>(A, Row, Col, R0, R1, C0, C1) {}
      _array_iterator() {}
      _array_iterator(const _array_iterator& x)
	: _array_iterator_base<Tp>(x._array, x.row, x.col) {}


      reference operator*() //const 
	{
	  return _array_iterator_base<Tp>::point();
	}

  
      self& operator++()
	{
	  this->_incr();
	  return *this;
	}

      self operator++(int)
	{
	  self tmp = *this;
	  this->_incr();
	  return tmp;
	}

      self& operator--()
	{
	  this->_decr();
	  return *this;
	}

      self operator--(int)
	{
	  self tmp = *this;
	  this->_decr();
	  return tmp;
	}

    };
}



#endif
