#include         <arrayTools/Convolution.h>

namespace _ARRAYTOOLS_NAMESPACE
{

//! Return the convolution of A by K 
//! The outer frame (whose size is determined by
//! that of the convolution kernel K) is filled with 0
//!

array<float>& Convolution(array<float>& A, array<float>& K)
 {
   array<float> *R;
   int		 x, y, wx, wy;
   int		 i, j;
   int           x0, x1;
   int           y0, y1;
   double	 v;
   float         **_A, **_K, *ptr, *_a;
   
   wx 		= K.cols() / 2;
   wy 		= K.rows() / 2;

   x0           = A.first_col_index() + wx;
   x1           = A.last_col_index()  - wx;
   y0           = A.first_row_index() + wy;
   y1           = A.last_row_index()  - wy;
   
   _A           = A.storage();
   _K           = K.storage();
   
   R 		= new array<float>(A.first_row_index(),
				   A.last_row_index(),
				   A.first_col_index(),
				   A.last_col_index());
   
   for(y = y0; y <= y1; y++)
      {
	for(x = x0; x <= x1; x++)
	   {
	     v = 0.0;
	     for(j = -wy; j <= wy; j++)
		{
		  ptr = _K[j+wy];
		  _a  = &(_A[y+j][x-wx]);

		  for(i = -wx;  i <= wx; i++)
		     {
		       v += (*_a++) * (*ptr++);
		     }
		}
	     (*R)[y][x] = v;
	   }
      }

   DeleteIfTemporary(A);
   DeleteIfTemporary(K);
   
   return((*R).temporary());
 }

array<float>& Convolution(array<float>&  A,
			  array<float>&  K,
			  array<u_char>& F,
			  float          fillValue)
 {
   array<float> *R;
   int		 x, y, wx, wy;
   int		 i, j;
   int           x0, x1;
   int           y0, y1;
   double	 v;
   float         **_A, **_K, *ptr, *_a;
   u_char        **_F;
   
   wx 		= K.cols() / 2;
   wy 		= K.rows() / 2;

   x0           = A.first_col_index() + wx;
   x1           = A.last_col_index()  - wx;
   y0           = A.first_row_index() + wy;
   y1           = A.last_row_index()  - wy;
   
   _A           = A.storage();
   _K           = K.storage();
   _F           = F.storage();
   
   R 		= new array<float>(A.first_row_index(),
				   A.last_row_index(),
				   A.first_col_index(),
				   A.last_col_index());
   
   for(y = y0; y <= y1; y++)
      {
	for(x = x0; x <= x1; x++)
	   {
	     if(_F[y][x])
		{
		  v = 0.0;
		  for(j = -wy; j <= wy; j++)
		     {
		       ptr = _K[j+wy];
		       _a  = &(_A[y+j][x-wx]);

		       for(i = -wx;  i <= wx; i++)
			  {
			    v += (*_a++) * (*ptr++);
			  }
		     }
		  (*R)[y][x] = v;
		}
	     else
		{
		  (*R)[y][x] = fillValue;
		} 
	   }
	
      }

   DeleteIfTemporary(A);
   DeleteIfTemporary(K);
   
   return((*R).temporary());
 }

}
