#include<arrayTools/Array.h>
#include<arrayTools/Utils.h>

#include<arrayTools/PngIO.h>

#include <png.h>
#include <assert.h>


/*
color_type     - describes which color/alpha
                     channels are present.
                     PNG_COLOR_TYPE_GRAY
                        (bit depths 1, 2, 4, 8, 16)
                     PNG_COLOR_TYPE_GRAY_ALPHA
                        (bit depths 8, 16)
                     PNG_COLOR_TYPE_PALETTE
                        (bit depths 1, 2, 4, 8)
                     PNG_COLOR_TYPE_RGB
                        (bit_depths 8, 16)
                     PNG_COLOR_TYPE_RGB_ALPHA
                        (bit_depths 8, 16)

                     PNG_COLOR_MASK_PALETTE
                     PNG_COLOR_MASK_COLOR
                     PNG_COLOR_MASK_ALPHA

interlace_type -     PNG_INTERLACE_NONE or
                     PNG_INTERLACE_ADAM7

*/

#define PNG_BYTES_TO_CHECK 4

/*
#ifndef png_jmpbuf
#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#endif
*/

namespace _ARRAYTOOLS_NAMESPACE
{

  
  int pngbuffer2array(unsigned char * buffer,
					  int rows, int cols, int color_type,
					  array<unsigned char>& I)
  {
	if( buffer == NULL)
      {
		return FALSE;
      }
    
    I.reset(rows,cols);

    int i,j;
    unsigned char ** _I = I.storage();
    unsigned char * __I;
    unsigned char * ptr = buffer;

    switch(color_type)
      {
      case PNG_COLOR_TYPE_GRAY:
		{
		  for(i = 0; i < rows; i++)
			{
			  __I = *_I++;
			  for(j = 0; j < cols; j++)
				{
				  *__I++ = *ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_GRAY_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __I = *_I++;
			  for(j = 0; j < cols; j++)
				{
				  *__I++ = *ptr++;
				  ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB:
		{
		  for(i = 0; i < rows; i++)
			{
			  __I = *_I++;
			  for(j = 0; j < cols; j++)
				{
				  *__I++ = (unsigned char)((ptr[0] + ptr[1] + ptr[2]) / 3);
				  ptr += 3;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __I = *_I++;
			  for(j = 0; j < cols; j++)
				{
				  *__I++ = (unsigned char)((ptr[0] + ptr[1] + ptr[2]) / 3);
				  ptr += 4;
				}
			}
		}
		break;
      default:
		fprintf(stderr,"PngIO : color type unrecognized\n");
		assert(0);
      }
    
    return TRUE;
  }

  int pngbuffer2rgb(unsigned char * buffer,
					int rows, int cols, int color_type,
					array<unsigned char>& R, 
					array<unsigned char>& G,
					array<unsigned char>& B)
  {
	if( buffer == NULL)
      {
		return FALSE;
      }
    
    R.reset(rows,cols);
    G.reset(rows,cols);
    B.reset(rows,cols);

    int i,j;
    unsigned char ** _R = R.storage();
    unsigned char ** _G = G.storage();
    unsigned char ** _B = B.storage();
    unsigned char * __R;
    unsigned char * __G;
    unsigned char * __B;
    unsigned char * ptr = buffer;

    switch(color_type)
      {
      case PNG_COLOR_TYPE_GRAY:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr;
				  *__G++ = *ptr;
				  *__B++ = *ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_GRAY_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr;
				  *__G++ = *ptr;
				  *__B++ = *ptr++;
				  ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr++;
				  *__G++ = *ptr++;
				  *__B++ = *ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr++;
				  *__G++ = *ptr++;
				  *__B++ = *ptr++;
				  ptr++;
				}
			}
		}
		break;
      default:
		fprintf(stderr,"PngIO : color type unrecognized\n");
		assert(0);
      }
    
    return TRUE;
  }

  int pngbuffer2rgba(unsigned char * buffer,
					 int rows, int cols, int color_type,
					 array<unsigned char>& R, 
					 array<unsigned char>& G,
					 array<unsigned char>& B,
					 array<unsigned char>& A)
  {
    if( buffer == NULL)
      {
		return FALSE;
      }
    
    R.reset(rows,cols);
    G.reset(rows,cols);
    B.reset(rows,cols);
    A.reset(rows,cols);


    int i,j;
    unsigned char ** _R = R.storage();
    unsigned char ** _G = G.storage();
    unsigned char ** _B = B.storage();
    unsigned char ** _A = A.storage();
    unsigned char * __R;
    unsigned char * __G;
    unsigned char * __B;
    unsigned char * __A;

    unsigned char * ptr = buffer;

    switch(color_type)
      {
      case PNG_COLOR_TYPE_GRAY:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  __A = *_A++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr;
				  *__G++ = *ptr;
				  *__B++ = *ptr++;
				  *__A++ = 255;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_GRAY_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  __A = *_A++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr;
				  *__G++ = *ptr;
				  *__B++ = *ptr++;
				  *__A++ = *ptr++;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  __A = *_A++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr++;
				  *__G++ = *ptr++;
				  *__B++ = *ptr++;
				  *__A++ = 255;
				}
			}
		}
		break;
      case PNG_COLOR_TYPE_RGB_ALPHA:
		{
		  for(i = 0; i < rows; i++)
			{
			  __R = *_R++;
			  __G = *_G++;
			  __B = *_B++;
			  __A = *_A++;
			  for(j = 0; j < cols; j++)
				{
				  *__R++ = *ptr++;
				  *__G++ = *ptr++;
				  *__B++ = *ptr++;
				  *__A++ = *ptr++;
				}
			}
		}
		break;
      default:
		fprintf(stderr,"PngIO : color type unrecognized\n");
		assert(0);
      }
    
    return TRUE;
  }


  int check_if_png(FILE *fp)
  {
    unsigned char buf[PNG_BYTES_TO_CHECK];

    if(fp == NULL) return FALSE;

    /* Read in some of the signature bytes */
    if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK)
      return FALSE;

    /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature.
       Return nonzero (true) if they match */

    return(!png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK));
  }

  int check_if_png(istream & s)
  {
    unsigned char buf[PNG_BYTES_TO_CHECK];

    s.read(reinterpret_cast<char*>(buf),PNG_BYTES_TO_CHECK);
  
    /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature.
       Return nonzero (true) if they match */

    return(!png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK));
  }


  //initialize the main struct without the io source
  int at_png_read_init(png_structp & png_ptr, png_infop & info_ptr)
  {
    
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
									 NULL, 
									 NULL, 
									 NULL);

    if (png_ptr == NULL)
      {
		return FALSE;
      }
    

    /* Allocate/initialize the memory for image information.  REQUIRED. */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL)
      {
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
		return FALSE;
      }

    return TRUE;
  }


  int at_png_write_init(png_structp & png_ptr, png_infop & info_ptr)
  {
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
									  NULL,
									  NULL,
									  NULL);
    if (png_ptr == NULL)
      {
		return FALSE;
      }
      
    info_ptr = png_create_info_struct(png_ptr);

    if (info_ptr == NULL)
      {
		png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
		return FALSE;
      }

    

    return TRUE;
  }


  //Main read function: is called after
  //at_png_read_init
  //and the setup of read functions
  unsigned char * at_png_read(png_structp png_ptr, png_infop info_ptr,
							  unsigned int sig_read, 
							  int & rows, int & cols, int & color_type, int & interlace_type)
  { 
    png_uint_32 width, height;
    int bit_depth;
    int row_bytes;
    int i;
    unsigned char * row_pointer;
    unsigned char * buffer;

    /* If we have already read some of the signature */
    png_set_sig_bytes(png_ptr, sig_read);
    
    png_read_info(png_ptr, info_ptr);

    png_get_IHDR(png_ptr, info_ptr, 
				 &width, &height, 
				 &bit_depth, &color_type,
				 &interlace_type, int_p_NULL, int_p_NULL);

    //Setup image tranformation
    //switch to 8bit

    /* transform paletted images into full-color rgb */
    if (color_type == PNG_COLOR_TYPE_PALETTE)
      png_set_palette_to_rgb(png_ptr);

    /* expand images to bit-depth 8 (only applicable for grayscale images) */
    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_gray_1_2_4_to_8(png_ptr);

    /* transform transparency maps into full alpha-channel */
    if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_tRNS_to_alpha(png_ptr);
    
    /* downgrade 16-bit images to 8 bit */
    if (bit_depth == 16)
      png_set_strip_16 (png_ptr);

    png_read_update_info (png_ptr, info_ptr);

    //reread the image info after the update
    png_get_IHDR(png_ptr, info_ptr, 
				 &width, &height, 
				 &bit_depth, &color_type,
				 &interlace_type, int_p_NULL, int_p_NULL);
    

    rows = height;
    cols = width;

    row_bytes = png_get_rowbytes (png_ptr, info_ptr);

    //fprintf(stderr,"r(%i) c(%i) bit_depth(%i) color_type(%i) row_bytes(%i)\n", rows,cols,bit_depth,color_type,row_bytes);

   
    //BUFFER allocation
    buffer = new unsigned char[row_bytes * height];

    if (buffer == NULL) 
      {
		png_destroy_read_struct (&png_ptr, &info_ptr, NULL);
		return NULL;
      }
    
    row_pointer = buffer; 

    for(i = 0; i < rows; i++)
      {
		png_read_row(png_ptr, row_pointer, NULL);
		row_pointer += row_bytes;
      }

    /* At this point you have read the entire image */

    png_read_end(png_ptr, info_ptr);

    /* clean up after the read, and free any memory allocated - REQUIRED */
    png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    
    
    /* that's it */
    return buffer;
  }



  //return the allocated buffer or NULL if fails

  unsigned char * at_png_read_filep(FILE * fp, int & rows, int & cols, int & color_type, int & interlace_type)
  {
    
    png_structp png_ptr = NULL;
    png_infop info_ptr = NULL;

    if(check_if_png(fp) == FALSE)
      {
		return NULL;
      }
    if(at_png_read_init(png_ptr,info_ptr) == FALSE)
      {
		return NULL;
      }

	if (setjmp((png_ptr->jmpbuf)))
      {
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    	return NULL;
      }

    /* Set up the input control if you are using standard C streams */
    png_init_io(png_ptr, fp);

    
    return at_png_read(png_ptr,info_ptr, PNG_BYTES_TO_CHECK, 
					   rows, cols, color_type, interlace_type);
  }


  void at_cppstream_read_fn(png_structp png_ptr,
							png_bytep data, png_size_t length)
  {
    ((istream*) png_ptr->io_ptr)->read(reinterpret_cast<char*>(data),length);
  }

  unsigned char * at_png_read_stream(istream & s, int & rows, int & cols, int & color_type, int & interlace_type)
  {
   
    png_structp png_ptr = NULL;
    png_infop info_ptr = NULL;
 
    if(check_if_png(s) == FALSE)
      {
		return NULL;
      }
    if(at_png_read_init(png_ptr,info_ptr) == FALSE)
      {
		return NULL;
      }

	if (setjmp((png_ptr->jmpbuf)))
      {
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    	return NULL;
      }

    png_set_read_fn(png_ptr, &s, at_cppstream_read_fn);

    return at_png_read(png_ptr, info_ptr, PNG_BYTES_TO_CHECK, 
					   rows, cols, color_type, interlace_type); 
  }


  //Main red function: is called after
  //at_png_write_init
  int at_png_write(png_structp png_ptr, png_infop info_ptr,
				   unsigned char * buffer,
				   png_uint_32 width, png_uint_32 height,
				   int bit_depth, int color_type)
  {
    /* set the zlib compression level */
    //png_set_compression_level(png_ptr,Z_BEST_COMPRESSION);

    /* set other zlib parameters */
    //png_set_compression_mem_level(png_ptr, 8);
    //png_set_compression_strategy(png_ptr,Z_DEFAULT_STRATEGY);
    //png_set_compression_window_bits(png_ptr, 15);
    //png_set_compression_method(png_ptr, 8);
    //png_set_compression_buffer_size(png_ptr, 8192);

    //setting header info
    
    png_set_IHDR(png_ptr, info_ptr, 
				 width, height,
				 bit_depth, color_type, 
				 PNG_INTERLACE_NONE,
				 PNG_COMPRESSION_TYPE_BASE,
				 PNG_FILTER_TYPE_BASE);

    //png_set_invert_alpha(png_ptr);

    png_write_info(png_ptr, info_ptr);

    //png_set_filler(png_ptr, 0, PNG_FILLER_BEFORE);
    png_set_packing(png_ptr);
    
    //png_set_bgr(png_ptr);

    //user transform
    //png_set_write_user_transform_fn(png_ptr,write_transform_fn);
    //png_set_user_transform_info(png_ptr, user_ptr, 0, 0);
    
    //png_write_flush(png_ptr);

    unsigned int i;
    unsigned char * row_pointer = buffer;
    int row_bytes   = png_get_rowbytes (png_ptr, info_ptr);

    for(i = 0; i < height; ++i)
      {
		png_write_row(png_ptr, row_pointer);
		row_pointer += row_bytes;
      }
    png_write_flush(png_ptr);

    png_write_end(png_ptr, info_ptr); 
    png_write_flush(png_ptr);

    //clean up
    png_destroy_write_struct(&png_ptr, &info_ptr);
    
    return TRUE;
  }


  int at_png_write_filep(FILE * fp, 
						 unsigned char * buffer,
						 png_uint_32 width, png_uint_32 height,
						 int bit_depth, int color_type)
  {
    png_structp png_ptr = NULL;
    png_infop info_ptr = NULL;

    if(at_png_write_init(png_ptr,info_ptr) == FALSE)
      {
		return FALSE;
      }

    png_init_io(png_ptr, fp);

    if (setjmp((png_ptr->jmpbuf)))
      {
		png_destroy_write_struct(&png_ptr, &info_ptr);
    	return FALSE;
      }

    return at_png_write(png_ptr, info_ptr,
						buffer,
						width, height,
						bit_depth, color_type);
		
  }


  void at_cppstream_write_fn(png_structp png_ptr,
							 png_bytep data, 
							 png_size_t length)
  {
    ((ostream*) png_ptr->io_ptr)->write(reinterpret_cast<char*>(data),length);
  }

  void at_cppstream_flush_fn(png_structp png_ptr)
  {
    ((ostream*) png_ptr->io_ptr)->flush();
  }

  int at_png_write_stream(ostream & o, 
						  unsigned char * buffer,
						  png_uint_32 width, png_uint_32 height,
						  int bit_depth, int color_type)
  {
    png_structp png_ptr = NULL;
    png_infop info_ptr = NULL;

    if(at_png_write_init(png_ptr,info_ptr) == FALSE)
      {
		return FALSE;
      }

    png_set_write_fn(png_ptr,&o, at_cppstream_write_fn,at_cppstream_flush_fn);

    if (setjmp((png_ptr->jmpbuf)))
      {
		png_destroy_write_struct(&png_ptr, &info_ptr);
    	return FALSE;
      }

    return at_png_write(png_ptr, info_ptr,
						buffer,
						width, height,
						bit_depth, color_type);
		
  }
  

  //===================================================================================== 
  //INTERFACE

  //=====================================================================================
  //READ

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  FILE * filestream)
  {
    if(filestream == NULL)
      {
		return FALSE;
      }
  
    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_filep(filestream, rows, cols, color_type, interlace_type);

    int out = pngbuffer2rgb(buffer,rows,cols,color_type,R,G,B);
      
    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  istream & s)
  {

    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_stream(s, rows, cols, color_type, interlace_type);

    int out = pngbuffer2rgb(buffer,rows,cols,color_type,R,G,B);

    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  const char * filename)
  {
    FILE * filestream = fopen (filename, "r");
    int out =  ReadPng(R,G,B,filestream);
    fclose(filestream);
    return out;
  }

  int ReadPng(array<unsigned char>& I, 
			  FILE * filestream)
  {
    if(filestream == NULL)
      {
		return FALSE;
      }
  
    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_filep(filestream, rows, cols, color_type, interlace_type);

    int out = pngbuffer2array(buffer,rows,cols,color_type,I);
      
    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& I, 
			  istream & s)
  {

    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_stream(s, rows, cols, color_type, interlace_type);

    int out = pngbuffer2array(buffer,rows,cols,color_type,I);

    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& I, 
			  char * filename)
  {
    FILE * filestream = fopen (filename, "r");
    int out =  ReadPng(I,filestream);
    fclose(filestream);
    return out;
  }

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  array<unsigned char>& A,
			  FILE * filestream)
  {
    if(filestream == NULL)
      {
		return FALSE;
      }
  
    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_filep(filestream, rows, cols, color_type, interlace_type);

    int out = pngbuffer2rgba(buffer,rows,cols,color_type,R,G,B,A);
      
    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  array<unsigned char>& A,
			  istream & s)
  {

    int cols, rows, color_type, interlace_type;
    unsigned char * buffer = at_png_read_stream(s, rows, cols, color_type, interlace_type);

    int out = pngbuffer2rgba(buffer,rows,cols,color_type,R,G,B,A);

    delete buffer;

    return out;
  }

  int ReadPng(array<unsigned char>& R, 
			  array<unsigned char>& G,
			  array<unsigned char>& B,
			  array<unsigned char>& A,
			  const char * filename)
  {
    FILE * filestream = fopen (filename, "r");
    int out =  ReadPng(R,G,B,A,filestream);
    fclose(filestream);
    return out;
  }

  //=====================================================================================
  //WRITE

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   FILE * filestream)
  { 
    if(filestream == NULL)
      {
		return FALSE;
      }
    if(!AreConsistent(R,G) ||
       !AreConsistent(R,B) ||
       R.is_empty())
      {
		return FALSE;
      }
        
    unsigned char * buffer = new unsigned char[3*R.cols()*R.rows()];
    rgb2buffer(R,G,B,buffer);
    
    at_png_write_filep(filestream,buffer,R.cols(),R.rows(),8,PNG_COLOR_TYPE_RGB);
    
    delete buffer;
    return TRUE;
  }

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   char * filename)
  {
    FILE * filestream = fopen(filename,"w");
    int out =  WritePng(R,G,B,filestream);
    fclose(filestream);
    return out;
  }

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   ostream & o)
  {
    if(!AreConsistent(R,G) ||
       !AreConsistent(R,B) ||
       R.is_empty())
      {
		return FALSE;
      }
    
    unsigned char * buffer = new unsigned char[3*R.cols()*R.rows()];
    rgb2buffer(R,G,B,buffer);
    
    at_png_write_stream(o,buffer,R.cols(),R.rows(),8,PNG_COLOR_TYPE_RGB);
    
    delete buffer;
    return TRUE;
  }

  int WritePng(array<unsigned char> & I, 
			   FILE * filestream)
  {
    if(filestream == NULL)
      {
		return FALSE;
      }
        
    if(I.is_empty())
      {
		return FALSE;
      }

    unsigned char * buffer = new unsigned char[I.cols()*I.rows()];
    array2buffer(I,buffer);
    
    at_png_write_filep(filestream,buffer,I.cols(),I.rows(),8,PNG_COLOR_TYPE_GRAY);
    
    delete buffer;
    return TRUE;
    
  }

  int WritePng(array<unsigned char>& I, 
			   char * filename)
  {
    FILE * filestream = fopen(filename,"w");
    int out =  WritePng(I,filestream);
    fclose(filestream);
    return out;
  }


  int WritePng(array<unsigned char>& I, 
			   ostream & o)
  {
    if(I.is_empty())
      {
		return FALSE;
      }
    
    unsigned char * buffer = new unsigned char[I.cols()*I.rows()];
    array2buffer(I,buffer);
    
    at_png_write_stream(o,buffer,I.cols(),I.rows(),8,PNG_COLOR_TYPE_RGB);
    
    delete buffer;
    return TRUE;
  }

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   array<unsigned char>& A,
			   FILE * filestream)
  { 
    if(filestream == NULL)
      {
		return FALSE;
      }
    if(!AreConsistent(R,G) ||
       !AreConsistent(R,B) ||
       !AreConsistent(R,A) ||
       R.is_empty())
      {
		return FALSE;
      }
        
    unsigned char * buffer = new unsigned char[4*R.cols()*R.rows()];
    rgba2buffer(R,G,B,A,buffer);
    
    at_png_write_filep(filestream,buffer,R.cols(),R.rows(),8,PNG_COLOR_TYPE_RGB_ALPHA);
    
    delete buffer;
    return TRUE;
  }

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   array<unsigned char>& A,
			   char * filename)
  {
    FILE * filestream = fopen(filename,"w");
    int out =  WritePng(R,G,B,A,filestream);
    fclose(filestream);
    return out;
  }

  int WritePng(array<unsigned char>& R, 
			   array<unsigned char>& G,
			   array<unsigned char>& B,
			   array<unsigned char>& A,
			   ostream & o)
  {
    if(!AreConsistent(R,G) ||
       !AreConsistent(R,B) ||
       !AreConsistent(R,A) ||
       R.is_empty())
      {
		return FALSE;
      }
    
    unsigned char * buffer = new unsigned char[4*R.cols()*R.rows()];
    rgba2buffer(R,G,B,A,buffer);
    
    at_png_write_stream(o,buffer,R.cols(),R.rows(),8,PNG_COLOR_TYPE_RGB_ALPHA);
    
    delete buffer;
    return TRUE;
  }
  


}//namespace
