#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arrayTools/aniCommon.h>

extern "C"  {

  int Rint(double value)
  {
    int    sign = 1;
    int    i_value;

    if(value < 0.0)
      {
		value = - value;
		sign  = -1;
      }
   
    i_value = int(value);

    if((value - i_value) > 0.5)
      i_value++;
   
    return(sign * i_value);
  }

  double Lanczos(double x,
				 double r)
  {
	// support*sin(pi*r)*sin(pi*r/support)/(pi*pi*r*r)
	if((x < -r) || (x > r))
	  {
		return(0.0);
	  }
	if(x == 0) 
	  {
		return(1.0);
	  } 
	else
	  {
		double l = r * sin(M_PI * x) * sin(M_PI*x/r) / (M_PI * M_PI * x * x);

		return(l);
	  }
  }

  SEXP aniAffineReal(SEXP animage_1, // Must be an animage object
					 SEXP r_M)       // must be an 3x3 array  representing the affine transform
  {
	SEXP     ani_r;
	int      x1, y1, w1, h1;
	double   o1, *p1, *p2;

    int		 cols, rows;
    int		 X_start, X_end;
    int		 Y_start, Y_end;
    int		 x_start, x_end, x;
    int		 y_start, y_end, y;
    double	 ia, ib, ic, id;
    double	 x_min, x_max, y_min, y_max;
    double	 bounds[4][2];
    double	 determinant;
   
    // localize images 
	//
	x1 = GET_ANIMAGE_X(animage_1);
	y1 = GET_ANIMAGE_Y(animage_1);
	w1 = GET_ANIMAGE_W(animage_1);
	h1 = GET_ANIMAGE_H(animage_1);
	o1 = GET_ANIMAGE_OR(animage_1);
	p1 = GET_ANIMAGE_DR(animage_1);
	
	// get the pointer to the affine matrix data
	// remember that they are store in column major order
	//
	double *Mptr = REAL(r_M);

	double a = Mptr[0];
	double c = Mptr[1];
	double b = Mptr[3];
	double d = Mptr[4];
	
	ia = ib = ic = id = 0;

	// get translation data
	//
	double O[2];

	O[0] = Mptr[6];
	O[1] = Mptr[7];
   
    // Check if transform is singular
	//
	{
	  determinant	= a*d - b*c;

	  if(determinant == 0.0)
		{
		  error("Singular transform detected by aniAffine");
		}
	  else 
		{
		  // compute inverse transform
		  //
		  ia	=   (d / determinant);
		  ib	= - (b / determinant);
		  ic	= - (c / determinant);
		  id	=   (a / determinant);
		}
	}

	cols		    = w1;
    rows		    = h1;
    X_start         = x1;
    X_end           = x1+cols-1;
    Y_start         = y1;
    Y_end           = y1+rows-1;
   
    bounds[0][0]	= (double) X_start;
    bounds[0][1]	= (double) Y_start;
    bounds[1][0]	= bounds[0][0] + (double) cols - 1;
    bounds[1][1]	= bounds[0][1]; 
    bounds[2][0]	= bounds[0][0]; 
    bounds[2][1]	= bounds[0][1] + (double) rows - 1;
    bounds[3][0]	= bounds[1][0];
    bounds[3][1]	= bounds[2][1];

    // Compute new bounding box
    //
    {
      double	x, y;
      int	i;
      
	  x_min = x_max	= a * (bounds[0][0] + O[0]) + b * (bounds[0][1] + O[1]);
	  y_min = y_max	= c * (bounds[0][0] + O[0]) + d * (bounds[0][1] + O[1]);

      for(i = 1; i < 4; i++)
		{
		  x     = a * (bounds[i][0] + O[0]) + b * (bounds[i][1] + O[1]);
		  y		= c * (bounds[i][0] + O[0]) + d * (bounds[i][1] + O[1]);
		  x_min = min(x_min, x);
		  x_max	= max(x_max, x);
		  y_min	= min(y_min, y);
		  y_max	= max(y_max, y);
		}

      x_start	= (int) Rint(x_min);
      x_end	    = (int) Rint(x_max);
      y_start	= (int) Rint(y_min);
      y_end	    = (int) Rint(y_max);
    }
 
	int x2 = x_start;
	int y2 = y_start;
	int w2 = x_end - x_start + 1;
	int h2 = y_end - y_start + 1;

	// Create return animage
	//
	PROTECT(ani_r      = newAnimage(x2, y2, w2, h2, o1));

	p2 = GET_ANIMAGE_DR(ani_r);

	// create pointers to read image data in random access order
	// preserving offsets
	// remember that the first index will be the column index
	//
	double **_ptr1 = (new double * [w1]) - x1;
	double **_ptr2 = (new double * [w2]) - x2;

	_ptr1[x1] = p1 - y1;
	for(int x = x1+1; x < x1 + w1; x++)
	  _ptr1[x] = _ptr1[x-1] + h1;

	_ptr2[x2] = p2 - y2;
	for(int x = x2+1; x < x2 + w2; x++)
	  _ptr2[x] = _ptr2[x-1] + h2;

    // and fill it
    //
    {
	  int    r = 2; // Lanczos support
	  double lx[7];
	  double ly[7];
	  double a0, a1;
	  double bx, by;
	  int    xx0, xx1, yy0, yy1;

      for(y = y_start; y <= y_end; y++)
		{
		  a0 = ib * y - O[0];
		  a1 = id * y - O[1]; 

		  for(x = x_start; x <= x_end; x++)
			{
			  bx	= ia * x + a0;
			  by 	= ic * x + a1;


			  xx0 =  int(ceil(bx - r));
			  xx1 =  int(floor(bx + r));
			  yy0 =  int(ceil(by - r));
			  yy1 =  int(floor(by + r));

			  // we want to keep within the support of the original
			  // image
			  //
			  xx0 = max(xx0, X_start);
			  yy0 = max(yy0, Y_start);
			  xx1 = min(xx1, X_end);
			  yy1 = min(yy1, Y_end);
			  
			  // we exploit Lanczos separability so we compute
			  // separately the horizontal/vertical weights
			  //
			  for(int i = xx0, j = 0; i <= xx1; i++, j++)
				lx[j] = Lanczos(i - bx, r);

			  for(int i = yy0, j = 0; i <= yy1; i++, j++)
				ly[j] = Lanczos(i - by, r);

			  double ws = 0.0, w;
			  double v  = 0.0;

			  for(int j = yy0, j1 = 0; j <= yy1; j++, j1++)
				for(int i = xx0, i1 = 0; i <= xx1; i++, i1++)
				  {
					w   = lx[i1] * ly[j1];
					ws += w;
					v  += w * _ptr1[i][j];
				  }
			  
			  if(ws > 0.0)
				_ptr2[x][y] = v / ws;
			  else
				_ptr2[x][y] = o1;
			}
		}
    }   

	// release side vectors
	//
	free(_ptr1+x1);
	free(_ptr2+x2);

	UNPROTECT(1);

	return(ani_r);
  }
}
