#include <arrayTools/aniCommon.h>


struct opdef optab[1024] = {
  {	"+",		PLUS},
  {	"-",		MINUS},
  {	"*",		TIMES},
  {	"/",		DIVIDE},
  {	"MIN",		MIN},
  {	"MAX",		MAX},
  {	"<=",		LESS_OR_EQ},
  {	"<",		LESS},
  {	">=",		GREATER_OR_EQ},
  {	">",		GREATER},
  {	"=",		EQUAL},
  {	"&",		AND},
  {	"|",		OR},
  {	"^",		XOR},
  {	"!",		NOT},
  {	"<<",		SHIFT_LEFT},
  {	">>",		SHIFT_RIGHT},
  {	"SET",		SET},
  {	NULL,	    NA},
};
		
int  statOpCode(const char *opName)
{
  int i = 0;

  while(optab[i].op_name != NULL) 
	{
	  
	  if(strcmp(opName, optab[i].op_name) == 0)
		return(optab[i].op_code);
	  else
		i++;
	}

  return(optab[i].op_code);
}
 
// Utility function to sort ROI for operation
//
void sorting(int *vector)
{
  int	    i, j, k, tmp;
    
  for(i = 0; i < 6; i++) {
	k = i;

	for(j = i; j < 6; j++) 
	  if(vector[j] < vector[k])
		k = j;

	tmp       = vector[i];
	vector[i] = vector[k];
	vector[k] = tmp;
  }    
}    

void statAni(SEXP ani,
			 SEXP *slot)
{
  
  slot[DATA]     = GET_SLOT(ani, install("data"));
  slot[MODE]     = GET_SLOT(ani, install("mode"));
  slot[FOCUS]    = GET_SLOT(ani, install("focus"));
  slot[OUTSIDE]  = GET_SLOT(ani, install("outside"));
  slot[STORAGE]  = GET_SLOT(ani, install("storage"));
  slot[DOCUMENT] = GET_SLOT(ani, install("document"));
  slot[HISTORY]  = GET_SLOT(ani, install("history"));
  
}
			 
SEXP newAnidata(int        w,
				int        h,
				aniStorage kind)
{
  SEXP ani_x = R_NilValue, dim;

  switch(kind)
	{
	case ANI_INTEGER:
	  PROTECT(ani_x = allocVector(INTSXP,  w * h));

	  {
		int *p = INTEGER(ani_x);
		int  n = w * h;
		while(n--) {
		  *p++ = 0;
		}
	  }
	  break;
	case ANI_REAL:
	  PROTECT(ani_x = allocVector(REALSXP, w * h));

	  {
		double *p = REAL(ani_x);
		int     n = w * h;

		while(n--) {
		  *p++ = 0.0;
		}

	  }
	  break;
	}
  
  PROTECT(dim = allocVector(INTSXP, 2));
  INTEGER(dim)[0] = h;
  INTEGER(dim)[1] = w;
  setAttrib(ani_x, R_DimSymbol, dim);
  UNPROTECT(2);

  return(ani_x);
}

SEXP newAnifocus(int x,
				 int y)
{
  SEXP ani_x;

  PROTECT(ani_x = allocVector(INTSXP, 2));

  INTEGER(ani_x)[0] = x;
  INTEGER(ani_x)[1] = y;

  UNPROTECT(1);

  return(ani_x);
}
				
SEXP newAnimask(int x,
				int y,
				int w,
				int h)
{
  SEXP ani_x;

  PROTECT(ani_x = NEW_OBJECT(MAKE_CLASS("animask")));

  SET_SLOT(ani_x, install("x"), ScalarInteger(x));
  SET_SLOT(ani_x, install("y"), ScalarInteger(y));
  SET_SLOT(ani_x, install("w"), ScalarInteger(w));
  SET_SLOT(ani_x, install("h"), ScalarInteger(h));
  
  UNPROTECT(1);

  return(ani_x);
}	

SEXP newAnimage(int        x,
				int        y,
				int        w,
				int        h,
				int        o)
{
  SEXP ani_r = R_NilValue;

  PROTECT(ani_r = NEW_OBJECT(MAKE_CLASS("animage")));

  SET_SLOT(ani_r, install("data"), newAnidata(w, h, ANI_INTEGER));
  SET_SLOT(ani_r, install("outside"), ScalarInteger(o));
  SET_SLOT(ani_r, install("storage"), allocVector(STRSXP, 1));

  SET_STRING_ELT(GET_SLOT(ani_r, install("storage")), 0, mkChar("integer"));

  GET_ANIMAGE_X(ani_r)  = x;
  GET_ANIMAGE_Y(ani_r)  = y;
  
  UNPROTECT(1);

  return(ani_r);
}

SEXP newAnimage(int        x,
				int        y,
				int        w,
				int        h,
				double     o)
{
  SEXP ani_r = R_NilValue;

  PROTECT(ani_r = NEW_OBJECT(MAKE_CLASS("animage")));

  SET_SLOT(ani_r, install("data"), newAnidata(w, h, ANI_REAL));
  SET_SLOT(ani_r, install("outside"), ScalarReal(o));
  SET_SLOT(ani_r, install("storage"), allocVector(STRSXP, 1));

  SET_STRING_ELT(GET_SLOT(ani_r, install("storage")), 0, mkChar("real"));

  GET_ANIMAGE_X(ani_r) = x;
  GET_ANIMAGE_Y(ani_r) = y;
  
  UNPROTECT(1);

  return(ani_r);
}

// A set of functions to quickly map arrays from R to arrayTools 
// and viceversa
//
SEXP ia2r_bi(ia::array<u_char> &A)
{
  SEXP ani_r = R_NilValue;
  int  x0, x1, y0, y1;

  A.stat(x0, x1, y0, y1);

  PROTECT(ani_r = newAnimage(x0, y0, x1-x0+1, y1-y0+1, int(0)));

  int     *p = GET_ANIMAGE_DI(ani_r);
  u_char **a = A.storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  *p++ = int(a[y][x]);

  UNPROTECT(1);

  return(ani_r);
}

SEXP ia2r_ii(ia::array<int> &A)
{
  SEXP ani_r = R_NilValue;
  int  x0, x1, y0, y1;

  A.stat(x0, x1, y0, y1);

  PROTECT(ani_r = newAnimage(x0, y0, x1-x0+1, y1-y0+1, int(0)));

  int    *p = GET_ANIMAGE_DI(ani_r);
  int   **a = A.storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  *p++ = int(a[y][x]);

  UNPROTECT(1);

  return(ani_r);
}

SEXP ia2r_fi(ia::array<float> &A)
{
  SEXP ani_r = R_NilValue;
  int  x0, x1, y0, y1;

  A.stat(x0, x1, y0, y1);

  PROTECT(ani_r = newAnimage(x0, y0, x1-x0+1, y1-y0+1, int(0)));

  int    *p = GET_ANIMAGE_DI(ani_r);
  float **a = A.storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  *p++ = int(a[y][x]);

  UNPROTECT(1);

  return(ani_r);
}

SEXP ia2r_di(ia::array<double> &A)
{
  SEXP ani_r = R_NilValue;
  int  x0, x1, y0, y1;

  A.stat(x0, x1, y0, y1);

  PROTECT(ani_r = newAnimage(x0, y0, x1-x0+1, y1-y0+1, int(0)));

  int     *p = GET_ANIMAGE_DI(ani_r);
  double **a = A.storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  *p++ = int(a[y][x]);

  DeleteIfTemporary(A);
  UNPROTECT(1);

  return(ani_r);
}

SEXP ia2r_dr(ia::array<double> &A)
{
  SEXP ani_r = R_NilValue;
  int  x0, x1, y0, y1;

  A.stat(x0, x1, y0, y1);

  PROTECT(ani_r = newAnimage(x0, y0, x1-x0+1, y1-y0+1, double(0)));

  double  *p = GET_ANIMAGE_DR(ani_r);
  double **a = A.storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  *p++ = double(a[y][x]);

  DeleteIfTemporary(A);
  UNPROTECT(1);

  return(ani_r);
}

ia::array<int> &r2ia_ii(SEXP animage)
{
  
  int  x0, x1, y0, y1, w, h;

  // localize images 
  //
  x0 = GET_ANIMAGE_X(animage);
  y0 = GET_ANIMAGE_Y(animage);
  w  = GET_ANIMAGE_W(animage);
  h  = GET_ANIMAGE_H(animage);

  x1 = x0 + w - 1;
  y1 = y0 + h - 1;

  ia::array<int> *A = (new ia::array<int>(y0,y1,x0,x1));

  int  *p = GET_ANIMAGE_DI(animage);
  int **a = (*A).storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  a[y][x] = *p++;

  return((*A).temporary());
}

ia::array<double> &r2ia_rd(SEXP animage)
{
  
  int  x0, x1, y0, y1, w, h;

  // localize images 
  //
  x0 = GET_ANIMAGE_X(animage);
  y0 = GET_ANIMAGE_Y(animage);
  w  = GET_ANIMAGE_W(animage);
  h  = GET_ANIMAGE_H(animage);

  x1 = x0 + w - 1;
  y1 = y0 + h - 1;

  ia::array<double> *A = (new ia::array<double>(y0,y1,x0,x1));

  double  *p = GET_ANIMAGE_DR(animage);
  double **a = (*A).storage();

  for(int x = x0; x <= x1; x++)
	for(int y = y0; y <= y1; y++)
	  a[y][x] = *p++;

  return((*A).temporary());
}

double min(double a, double b)
{
  return((a < b) ? a : b);
}

double max(double a, double b)
{
  return((a > b) ? a : b);
}

int min(int a, int b)
{
  return((a < b) ? a : b);
}

int max(int a, int b)
{
  return((a > b) ? a : b);
}

