#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arrayTools/aniCommon.h>

extern "C"  {

  SEXP aniConvolution(SEXP animage_1, // Must be an animage object
					  SEXP kernel)    // must be an array with odd dimensions
  {
	SEXP ani_r;
	int  x1, y1, w1, h1, o1, *p1;

    // localize images 
	//
	x1 = GET_ANIMAGE_X(animage_1);
	y1 = GET_ANIMAGE_Y(animage_1);
	w1 = GET_ANIMAGE_W(animage_1);
	h1 = GET_ANIMAGE_H(animage_1);
	o1 = GET_ANIMAGE_OI(animage_1);
	p1 = GET_ANIMAGE_DI(animage_1);
	
	// Create return animage
	//
	PROTECT(ani_r      = newAnimage(x1, y1, w1, h1, o1));

	// Get info on the kernel
	//
	int  wk = INTEGER(GET_DIM(kernel))[1];
	int  hk = INTEGER(GET_DIM(kernel))[0];
	int  h  = hk/2;
	int  w  = wk/2;
	int  of = h1 - hk;

	int *sraptr = p1 - (w*h1 + h);
	int *srkptr = INTEGER(kernel);
	int *waptr  = GET_ANIMAGE_DI(ani_r);

	for(int x = 0; x < w1; x++) 
	  {
		for(int y = 0; y < h1; y++)
		  {
			// reset pointers 
			//
			int *raptr = sraptr;
			int *rkptr = srkptr;
			int  sum   = 0;

			for(int xx = -w; xx <= w; xx++)
			  {
				int boolx = (x+xx >= 0) && (x+xx < w1);

				for(int yy = -h; yy <= h; yy++)
				  {
					if(boolx && (y+yy >= 0) && (y+yy < h1)) 
					  sum += (*raptr) * (*(rkptr++));
					else
					  sum += o1       * (*(rkptr++));
					raptr++;
				  }
				
				raptr += of;
			  }

			*(waptr++) = sum;
			sraptr++;
		  }
	  }
	
	UNPROTECT(1);

	return(ani_r);
  }

  SEXP aniConvolutionReal(SEXP animage_1, // Must be an animage object
						  SEXP kernel)    // must be an array with odd dimensions
  {
	SEXP   ani_r;
	int    x1, y1, w1, h1;
	double o1, *p1;

    // localize images 
	//
	x1 = GET_ANIMAGE_X(animage_1);
	y1 = GET_ANIMAGE_Y(animage_1);
	w1 = GET_ANIMAGE_W(animage_1);
	h1 = GET_ANIMAGE_H(animage_1);
	o1 = GET_ANIMAGE_OR(animage_1);
	p1 = GET_ANIMAGE_DR(animage_1);
	
	// Create return animage
	//
	PROTECT(ani_r      = newAnimage(x1, y1, w1, h1, o1));

	// Get info on the kernel
	//
	int  wk = INTEGER(GET_DIM(kernel))[1];
	int  hk = INTEGER(GET_DIM(kernel))[0];
	int  h  = hk/2;
	int  w  = wk/2;
	int  of = h1 - hk;

	double *sraptr = p1 - (w*h1 + h);
	double *srkptr = REAL(kernel);
	double *waptr  = GET_ANIMAGE_DR(ani_r);

	for(int x = 0; x < w1; x++) 
	  {
		for(int y = 0; y < h1; y++)
		  {
			// reset pointers 
			//
			double *raptr = sraptr;
			double *rkptr = srkptr;
			double  sum   = 0;

			for(int xx = -w; xx <= w; xx++)
			  {
				int boolx = (x+xx >= 0) && (x+xx < w1);

				for(int yy = -h; yy <= h; yy++)
				  {
					if(boolx && (y+yy >= 0) && (y+yy < h1)) 
					  sum += (*raptr) * (*(rkptr++));
					else
					  sum += o1       * (*(rkptr++));
					raptr++;
				  }
				
				raptr += of;
			  }

			*(waptr++) = sum;
			sraptr++;
		  }
	  }
	
	UNPROTECT(1);

	return(ani_r);
  }
}
