#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <tiffio.h>
#include <arrayTools/TiffIO.h>
#include <arrayTools/PngIO.h>


#include <arrayTools/aniCommon.h>

extern "C"  {
  // Return the number of channels in a tiff image
  //
  SEXP at_tiffstat(SEXP fileName) 
  {
	const char *name = CHAR(STRING_ELT(fileName, 0));
	TIFF* tif  = TIFFOpen(name, "r"); 
	
	if(tif) 
	  { 
		uint16    sampleformat    = 0;
		uint16    samplesperpixel = 0;
		uint16    bitspersample   = 0;
		uint16    extrasamples    = 0;
		uint16   *extrasamplesInfos;
		uint32    height          = 0; 
		uint32    width           = 0; 
		SEXP      info;

		TIFFGetField(tif, TIFFTAG_SAMPLEFORMAT,    &sampleformat); 
		TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel); 
		TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   &bitspersample); 
		TIFFGetField(tif, TIFFTAG_EXTRASAMPLES,    &extrasamples, &extrasamplesInfos);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     &height); 
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      &width); 

		PROTECT(info = NEW_LIST(6));
		SET_ELEMENT(info, 0, ScalarInteger(int(sampleformat)));
		SET_ELEMENT(info, 1, ScalarInteger(int(samplesperpixel)));
		SET_ELEMENT(info, 2, ScalarInteger(int(bitspersample)));
		SET_ELEMENT(info, 3, ScalarInteger(int(extrasamples)));
		SET_ELEMENT(info, 4, ScalarInteger(int(height)));
		SET_ELEMENT(info, 5, ScalarInteger(int(width)));
		UNPROTECT(1);

		return(info);
	  }
	else
	  return(R_NilValue);
  }

  // Read a  tiff image 
  //
  SEXP at_readtiff(SEXP fileName) 
  {
	const char *name = CHAR(STRING_ELT(fileName, 0));
	TIFF* tif  = TIFFOpen(name, "r"); 
  
	if(tif) 
	  { 
		uint16    sampleformat    = 0;
		uint16    samplesperpixel = 0;
		uint16    bitspersample   = 0;
		uint16    extrasamples    = 0;
		uint16   *extrasamplesInfos;
		uint32    height          = 0; 
		uint32    width           = 0; 
		tdata_t   buf; 
		uint32    row; 
		SEXP      ans;
		SEXP      channels;

		TIFFGetField(tif, TIFFTAG_SAMPLEFORMAT,    &sampleformat); 
		TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel); 
		TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   &bitspersample); 
		TIFFGetField(tif, TIFFTAG_EXTRASAMPLES,    &extrasamples, &extrasamplesInfos);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     &height); 
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      &width); 
		
		int    nch              = samplesperpixel;

		switch(sampleformat) {
		case 0:
		case SAMPLEFORMAT_UINT:
		  switch(bitspersample) {
		  case 8:
			{
			  ia::array<u_char>  Ch[samplesperpixel];
			  u_char *           p;

			  for(int c = 0; c < nch; c++)
				Ch[c] = (* new ia::array<u_char> (height,  width)).temporary();

			  buf = _TIFFmalloc(TIFFScanlineSize(tif)); 
			  p   = (u_char *) buf;

			  for (row = 0; row < height; row++) 
				{
				  TIFFReadScanline(tif, buf, row);
	  
				  for(uint32 x = 0; x < width; x++)
					for(int c = 0; c < nch; c++)
					  Ch[c][row][x] = (*p++);

				  p   = (u_char *) buf;
				}

			  _TIFFfree(buf); 
			  TIFFClose(tif); 

			  PROTECT(channels = NEW_LIST(nch));
			  
			  for(int c = 0; c < nch; c++)
				{
				  SET_ELEMENT(channels, c, (ans = allocMatrix(INTSXP, height, width)));
				  
				  int *ians = INTEGER(ans);
				  
				  for(uint32 y = 0; y < height; y++)
					for(uint32 x = 0; x < width; x++)
					  ians[height*x+y] = Ch[c][y][x];
				}
			  
			  UNPROTECT(1);

			  return(channels);
			}
			break;
		  case 16:
			{
			  ia::array<u_short> Ch[samplesperpixel];
			  u_short *           p;

			  for(int c = 0; c < nch; c++)
				Ch[c] = (* new ia::array<u_short> (height,  width)).temporary();

			  buf = _TIFFmalloc(TIFFScanlineSize(tif)); 
			  p   = (u_short *) buf;

			  for (row = 0; row < height; row++) 
				{
				  TIFFReadScanline(tif, buf, row);
	  
				  for(uint32 x = 0; x < width; x++)
					for(int c = 0; c < nch; c++)
					  Ch[c][row][x] = (*p++);

				  p   = (u_short *) buf;
				}

			  _TIFFfree(buf); 
			  TIFFClose(tif); 

			  PROTECT(channels = NEW_LIST(nch));
			  
			  for(int c = 0; c < nch; c++)
				{
				  SET_ELEMENT(channels, c, (ans = allocMatrix(INTSXP, height, width)));
				  
				  int *ians = INTEGER(ans);
				  
				  for(uint32 y = 0; y < height; y++)
					for(uint32 x = 0; x < width; x++)
					  ians[height*x+y] = Ch[c][y][x];
				}
			  
			  UNPROTECT(1);

			  return(channels);
			}
			break;
		  default:
			return(R_NilValue);
		  }
		  break;
		case SAMPLEFORMAT_IEEEFP:
		  {
			ia::array<float> Ch[samplesperpixel];
			float *           p;

			for(int c = 0; c < nch; c++)
			  Ch[c] = (* new ia::array<float> (height,  width)).temporary();

			buf = _TIFFmalloc(TIFFScanlineSize(tif)); 
			p   = (float *) buf;

			for (row = 0; row < height; row++) 
			  {
				TIFFReadScanline(tif, buf, row);
	  
				for(uint32 x = 0; x < width; x++)
				  for(int c = 0; c < nch; c++)
					Ch[c][row][x] = (*p++);

				p   = (float *) buf;
			  }

			_TIFFfree(buf); 
			TIFFClose(tif); 

			PROTECT(channels = NEW_LIST(nch));
			  
			for(int c = 0; c < nch; c++)
			  {
				SET_ELEMENT(channels, c, (ans = allocMatrix(REALSXP, height, width)));
				  
				double *rans = REAL(ans);
				  
				for(uint32 y = 0; y < height; y++)
				  for(uint32 x = 0; x < width; x++)
					rans[height*x+y] = Ch[c][y][x];
			  }
			  
			UNPROTECT(1);

			return(channels);
		  }
		  break;
		default:
		  return(R_NilValue);
		}
	  }
	else
	  return(R_NilValue);
  }

  // Write a multichannel  tiff image 
  // If includeTransparency is TRUE
  // the last channel is  assumed to contain transparency information.
  // Values above 255/65535/1.0 are clipped.
  //
  // If dataformat is 2 images must be real
  //
  SEXP at_writetiff(SEXP imageChannels,
					SEXP hasTransparency,
					SEXP fileName, 
					SEXP dataFormat)  // 0: 8bit 1: 16bit 2:float

  {
	int     format          = INTEGER_VALUE(dataFormat);
	int     transparency    = (LOGICAL_VALUE(hasTransparency)) ? 1 : 0;
	const char   *name      = CHAR(STRING_ELT(fileName, 0));
	int     width           = GET_ANIMAGE_W(VECTOR_ELT(imageChannels, 0));
	int     height          = GET_ANIMAGE_H(VECTOR_ELT(imageChannels, 0));
	int     channels        = GET_LENGTH(imageChannels);
	int     bitspersample   = 0;
	TIFF   *out             = TIFFOpen(name, "w");

	switch(format) {
	case 0:
	  bitspersample = 8;
	  break;
	case 1:
	  bitspersample = 16;
	  break;
	case 2:
	  bitspersample = 32;
	  TIFFSetField(out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
	  break;
	}

	tsize_t linebytes       = channels * width * (bitspersample / 8); 
	char   *image           = new char [height*linebytes];

	TIFFSetField(out, TIFFTAG_IMAGEWIDTH, width);                     // set the width of the image
	TIFFSetField(out, TIFFTAG_IMAGELENGTH, height);                   // set the height of the image
	TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, channels);             // set number of channels per pixel
	TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, bitspersample);          // set the size of the channels
	TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);      // set the origin of the image.
	TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);

	if((channels - transparency) == 1) 
	  TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK); // Gray level image
	else if ((channels - transparency) == 3) 
	  TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);        // RGB image

	if(transparency > 0)
	  TIFFSetField(out, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_ASSOCALPHA);// set the size of the channels

	unsigned char *buf = (unsigned char *)_TIFFmalloc(TIFFScanlineSize(out));
  
	// We set the strip size of the file to be size of one row of pixels
	//
	TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, linebytes));
  

	switch(format){
	case 0:
	  {
		u_char  *p           = (u_char *) image;
		int    **channelData = new int *[channels];
	
		for(int c = 0; c < channels; c++) 
		  channelData[c] = GET_ANIMAGE_DI(VECTOR_ELT(imageChannels, c));
	
		for(int y = 0; y < height; y++)
		  for(int x = 0; x < width; x++)
			for(int c = 0; c < channels; c++) 
			  {
				int value = channelData[c][height*x+y];

				if(value < 0) 
				  value =   0;
				else if (value > 255)
				  value = 255;

				*p++ = value;
			  }

		delete channelData;
	  }
	  break;
	case 1:
	  {
		u_short  *p           = (u_short *) image;
		int     **channelData = new int *[channels];
		
		for(int c = 0; c < channels; c++) 
		  channelData[c] = GET_ANIMAGE_DI(VECTOR_ELT(imageChannels, c));
	
		for(int y = 0; y < height; y++)
		  for(int x = 0; x < width; x++)
			for(int c = 0; c < channels; c++) 
			  {
				int value = channelData[c][height*x+y];

				if(value < 0) 
				  value =   0;
				else if (value > 65535)
				  value = 65535;

				*p++ = value;
			  }

		delete channelData;
	  }
	  break;
	case 2:
	  {
		float   *p           = (float *) image;
		double **channelData = new double *[channels];
		
		for(int c = 0; c < channels; c++) 
		  channelData[c] = GET_ANIMAGE_DR(VECTOR_ELT(imageChannels, c));
	
		for(int y = 0; y < height; y++)
		  for(int x = 0; x < width; x++)
			for(int c = 0; c < channels; c++) 
			  {
				float value = channelData[c][height*x+y];

				*p++ = value;
			  }
		
		delete channelData;
	  }
	  break;
	}

	// Write image to the file one strip at a time
	//
	{
	  uint32 uheight = (uint32)height;

	  for (uint32 row = 0; row < uheight; row++)
		{
		  memcpy(buf, &image[row*linebytes], linebytes); 
		  if (TIFFWriteScanline(out, buf, row, 0) < 0)
			break;
		}
	}


	(void) TIFFClose(out);

	if (buf)
	  _TIFFfree(buf);


	// release temporary storage
	//
	delete image;

	return(ScalarLogical(true));
  }

  // Read a PNG image providing correct alpha
  // information 
  //
  SEXP at_readpng_rgba(SEXP fileName) 
  {
	ia::array<u_char>  RGBA[4];

	int               *ians;
	const char        *name;
	SEXP               ans;
	SEXP               channels;


	name = CHAR(STRING_ELT(fileName, 0));

	ia::ReadPng(RGBA[0],
				RGBA[1],
				RGBA[2], 
				RGBA[3], 
				name);

	int                w = RGBA[0].cols();
	int                h = RGBA[0].rows();

	PROTECT(channels = NEW_LIST(4));

	for(int ch = 0; ch < 4; ch++)
	  {
		SET_ELEMENT(channels, ch, (ans = allocMatrix(INTSXP, h, w)));

		ians = INTEGER(ans);

		for(int y = 0; y < h; y++)
		  for(int x = 0; x < w; x++)
			{
			  ians[h*x+y] = RGBA[ch][y][x];
			}
	  }

	UNPROTECT(1);

	return(channels);
  }

}
