#ifndef ARRAYTOOLS_H
#define ARRAYTOOLS_H

#include <arrayTools/Array.h>

#include <arrayTools/namespace_header.h>

namespace _ARRAYTOOLS_NAMESPACE
{

int   statArray(char *filename, int *r0, int *r1, int *c0, int *c1, int *type);

// Obsolete:
// void appendToFile(char *name, array<float> &A);
// void appendToFile(char *name, array<u_char> &A);



template<class T> void appendToFile(char * name, array<T> & A)
 {
   int           y, r0, r1;
   int           x, c0, c1;
   int           s0, s1, arrayType;
   int           recordsize = 0;
   int           file_size = 0;
   FILE          *bout = NULL;
   animal_header header;
   
   if(statArray(name, &s0, &s1, &c0, &c1, &arrayType) != 1)
      {
	//! if cannot stat it, create it. 
	//!
	A.write(name, AnimalType<T>() );
      }
   else
      {
	arrayType = AnimalType<T>();

	//! Open file
	//!
	if((bout = fopen(name, "r+b")) == NULL)
	   {
	     fprintf(stderr,
		     "appendToFile : Cannot open %s for appending data!\n",
		     name);
	     exit(-1);
	   }
	else
	   {
	     //! Check if the file was originally empty
	     //!
	     fseek(bout, 0 , SEEK_END);
	     file_size = ftell(bout);
	     if(file_size == 0)
		{
		  //! Prepare an empty header
		  //!
		  header.MAGIC	= lIMAGE;
		  header.rows	= 0;
		  header.cols	= A.cols();
		  header.type	= arrayType;
		  header.focusx	= c0 = A.first_col_index();
		  header.focusy	= A.first_row_index();
		  header.outside     = 0;
		  header.flag	= 0;
		  
		  c1                 = c0 + A.cols() - 1;
		}
	     else
		{
		  //! Position at 0
		  //!
		  fseek(bout, 0 , SEEK_SET);
		  
		  //! Read header
		  //!
		  freadx((int *) &header,
			 sizeof(int),
			 sizeof(animal_header)/sizeof(int), bout);   
		  
		  //! Check data type
		  //!
		  if(header.type != arrayType)
		     {
		       fprintf(stderr, "appendToFile : Data type mismatch!\n");
		       exit(-1);
		     }
		  else
		     {
		       //! check record  boundaries
		       //!
		       c0   = header.focusx;
		       c1   = c0 + header.cols - 1;
		       
		       if((c0 != A.first_col_index()) || (c1 != A.last_col_index()))
			  {
			    fprintf(stderr, "appendToFile : Cannot append to file, dimension mismatch!\n");
			    exit(-1);
			  }
		     }
		}
	   }
	
	//! Determine size of row
	//!
	recordsize  = header.cols * sizeof(T);
	
	//! Position at end of file
	//!
	fseek(bout, sizeof(animal_header) + header.rows * recordsize, SEEK_SET);
	
	//! and append data
	//!
	 {
	   T value;
	   T** _A = A.storage();
	   
	   r1 = A.last_row_index();
	   r0 = A.first_row_index();
	   
	   for(y = r0; y <= r1; y++)
	     for(x = c0; x <= c1; x++)
		{
		  value	= _A[y][x];
		  fwritex((char *) &value, sizeof(T), 1, bout);
		}
	 }
	
	//! Position at start of file and 
	//! write updated header
	//!
	header.rows += (r1 - r0 + 1);
	fseek(bout, 0 , SEEK_SET);
	fwritex((char *) &header,
		sizeof(int),
		sizeof(animal_header)/sizeof(int),
		bout);
	
	fclose(bout);
      }// if else

   DeleteIfTemporary(A);
 }

}


#endif
