#ifndef CORRELATION_H
#define  CORRELATION_H


#include <arrayTools/Array.h>

#include <arrayTools/namespace_header.h>


namespace _ARRAYTOOLS_NAMESPACE
{

array<float>& UNCorrelation(array<float>&  II,   // Image where the template is searched for
			    array<float>&  TT,   // Template to be located
			    int            dd,   // Size of neighborhood
			    int            kind, // can be 'S' 'G' or 'L'
			    array<float>*  F,    // Filter of image
			    array<float>*  M,    // Mask of template
			    int           *rx,   // Return x coordinate of best value
			    int           *ry,   // Return y coordinate of best value
			    float         *mr    // Return best value
			    );

array<float>& Correlation(array<float>&  II,   // Image where the template is searched for
			  array<float>&  TT,   // Template to be located
			  int            dd,   // Size of neighborhood
			  int            kind, // can be 'S' 'G' or 'L'
			  array<float>*  F,    // Filter of image
			  array<float>*  M,    // Mask of template
			  int           *rx,   // Return x coordinate of best value
			  int           *ry,   // Return y coordinate of best value
			  float         *mr    // Return best value
			  );


array<float>& HierarchicalCorrelation(array<float>* image_pyramid,
				      array<float>* template_pyramid,
				      int    levels,
				      int*   neighborhood,
				      int    kind, // can be 'S' 'G' or 'L'
				      double fraction_to_keep,
				      int    *rx,  // Return x coordinate of best value
				      int    *ry,  // Return y coordinate of best value
				      double *mr   // Return best value
				      );

array<float>& CorrelationW(array<float>&  II,   // Image where the template is searched for
			   array<float>&  TT,   // Template to be located
			   int            dd,   // Size of neighborhood
			   int            kind, // can be 'S' 'G' or 'L'
			   array<float>*  F,    // Filter of image
			   array<float>*  M,    // Mask of template
			   int           *rx,   // Return x coordinate of best value
			   int           *ry,   // Return y coordinate of best value
			   float         *mr    // Return best value
			   );



array<float>& HierarchicalCorrelationW(array<float>* image_pyramid,
				       array<float>* template_pyramid,
				       array<float>* mask_pyramid,
				       int    levels,
				       int*   neighborhood,
				       int    kind, // can be 'S' 'G' or 'L'
				       double fraction_to_keep,
				       int    *rx,  // Return x coordinate of best value
				       int    *ry,  // Return y coordinate of best value
				       double *mr,  // Return best value
				       vector<double> *bvs = NULL,
				       int     unusedLevels = 0
				       );

array<float>& HierarchicalCorrelation(array<float>* image_pyramid,
				      array<float>* template_pyramid,
				      array<float>* mask_pyramid,
				      int    levels,
				      int*   neighborhood,
				      int    kind, // can be 'S' 'G' or 'L'
				      double fraction_to_keep,
				      int    *rx,  // Return x coordinate of best value
				      int    *ry,  // Return y coordinate of best value
				      double *mr,  // Return best value
				      vector<double> *bvs = NULL,
				      int     unusedLevels = 0
				      );


array<double>& Correlation(array<double>&  II,   // Image where the template is searched for
			   array<double>&  TT,  // Template to be located
			   int            dd,   // Size of neighborhood
			   int            kind, // can be 'S' 'G' or 'L'
			   array<double>*  F,   // Filter of image
			   array<double>*  M,   // Mask of template
			   int           *rx,   // Return x coordinate of best value
			   int           *ry,   // Return y coordinate of best value
			   double         *mr   // Return best value
			   );


array<double>& HierarchicalCorrelation(array<double>* image_pyramid,
				      array<double>* template_pyramid,
				      int    levels,
				      int*   neighborhood,
				      int    kind, // can be 'S' 'G' or 'L'
				      double fraction_to_keep,
				      int    *rx,  // Return x coordinate of best value
				      int    *ry,  // Return y coordinate of best value
				      double *mr   // Return best value
				      );


double  RemapCorrelationValue(double value, int kind);
 typedef  double (*d_binop) (double A, double B);

typedef array<float>& (*GenCorrFunct)(array<float>* image_pyramid,
				      array<float>* template_pyramid,
				      array<float>* mask_pyramid,
				      int    levels,
				      int*   neighborhood,
				      int    kind, 
				      double fraction_to_keep,
				      int    *rx, 
				      int    *ry, 
				      double *mr,  
				      vector<double> *bvs,
				      int     unusedLevels);

float Correlation(array<float>& II,   // Image where the template is searched for
				  array<float>& TT);   // Template to be located

double Correlation(array<double>& II,   // Image where the template is searched for
				   array<double>& TT);   // Template to be located
 
 array<double>& CorrelationW(array<double>&  II,  // Image where the template is searched for
							 array<double>&  TT,  // Template to be located
							 int            dd,   // Size of neighborhood
							 int            kind, // can be 'S' 'G' or 'L'
							 array<double>*  F,   // Filter of image
							 array<double>*  M,   // Mask of template
							 int           *rx,   // Return x coordinate of best value
							 int           *ry,   // Return y coordinate of best value
							 double         *mr   // Return best value
							 );

 array<double>& CorrelationW2(array<double>&  II,   // Image where the template is searched for
							  array<double>&  TT,   // Template to be located
							  int             dd,   // Size of neighborhood
							  int             kind, // can be 'S' 'G' 'L' 'P'
							  int             normalize, // 0/1
							  int             p,    // used to determine L_p
							  double          maxD, // used to determine L_p
							  array<double>*  F,    // Filter of image
							  array<double>*  M,    // Mask of template
							  int            *rx,   // Best value x
							  int            *ry,   // Best value y
							  double         *mr    // Best value
							  );

 array<float>& CorrelationW2(array<float>&  II,         // Image where the template is searched for
							 array<float>&  TT,         // Template to be located
							 int             dd,        // Size of neighborhood
							 int             kind,      // can be 'S' 'G' 'L' 'P'
							 int             normalize, // 0/1
							 int             p,         // used to determine L_p
							 double          maxD,      // used to determine L_p
							 array<float>*  F,          // Filter of image
							 array<float>*  M,          // Mask of template
							 int            *rx,        // Best value x
							 int            *ry,        // Best value y
							 double         *mr         // Best value
							 );

  array<float>& HierarchicalCorrelationW2(array<float>*   image_pyramid,
										  array<float>*   template_pyramid,
										  array<float>*   filter_pyramid,
										  array<float>*   mask_pyramid,
										  int             levels,
										  int*            neighborhood,
										  int             kind, 
										  int             p,
										  double          maxD,
										  double          fraction_to_keep,
										  int            *rx, 
										  int            *ry, 
										  double         *mr,  
										  vector<double> *bvs,
										  int             unusedLevels
										  );
}



#endif
